from copy import deepcopy

import numpy as np

from ...util.SEBESOLWEIGCommonFiles.clearnessindex_2013b import clearnessindex_2013b
from ...util.SEBESOLWEIGCommonFiles.create_patches import create_patches
from ...util.SEBESOLWEIGCommonFiles.diffusefraction import diffusefraction
from ...util.SEBESOLWEIGCommonFiles.Perez_v3 import Perez_v3
from ...util.SEBESOLWEIGCommonFiles.shadowingfunction_wallheight_13 import shadowingfunction_wallheight_13
from ...util.SEBESOLWEIGCommonFiles.shadowingfunction_wallheight_23 import shadowingfunction_wallheight_23
from .anisotropic_sky import anisotropic_sky as ani_sky
from .cylindric_wedge import cylindric_wedge
from .daylen import daylen
from .gvf_2018a import gvf_2018a

# from .Lside_veg_v2015a import Lside_veg_v2015a
# from .Kside_veg_v2019a import Kside_veg_v2019a
from .Kside_veg_v2022a import Kside_veg_v2022a
from .Kup_veg_2015a import Kup_veg_2015a

# Anisotropic longwave
from .Lside_veg_v2022a import Lside_veg_v2022a
from .patch_radiation import patch_steradians
from .TsWaveDelay_2015a import TsWaveDelay_2015a

# Wall surface temperature scheme
from .wall_surface_temperature import wall_surface_temperature


def Solweig_2025a_calc(
    i,
    dsm,
    scale,
    rows,
    cols,
    svf,
    svfN,
    svfW,
    svfE,
    svfS,
    svfveg,
    svfNveg,
    svfEveg,
    svfSveg,
    svfWveg,
    svfaveg,
    svfEaveg,
    svfSaveg,
    svfWaveg,
    svfNaveg,
    vegdem,
    vegdem2,
    albedo_b,
    absK,
    absL,
    ewall,
    Fside,
    Fup,
    Fcyl,
    altitude,
    azimuth,
    zen,
    jday,
    usevegdem,
    onlyglobal,
    buildings,
    location,
    psi,
    landcover,
    lc_grid,
    dectime,
    altmax,
    dirwalls,
    walls,
    cyl,
    elvis,
    Ta,
    RH,
    radG,
    radD,
    radI,
    P,
    amaxvalue,
    bush,
    Twater,
    TgK,
    Tstart,
    alb_grid,
    emis_grid,
    TgK_wall,
    Tstart_wall,
    TmaxLST,
    TmaxLST_wall,
    first,
    second,
    svfalfa,
    svfbuveg,
    firstdaytime,
    timeadd,
    timestepdec,
    Tgmap1,
    Tgmap1E,
    Tgmap1S,
    Tgmap1W,
    Tgmap1N,
    CI,
    TgOut1,
    diffsh,
    shmat,
    vegshmat,
    vbshvegshmat,
    anisotropic_sky,
    asvf,
    patch_option,
    voxelMaps,
    voxelTable,
    ws,
    wallScheme,
    timeStep,
    steradians,
    walls_scheme,
    dirwalls_scheme,
):
    # def Solweig_2021a_calc(i, dsm, scale, rows, cols, svf, svfN, svfW, svfE, svfS, svfveg, svfNveg, svfEveg, svfSveg,
    #                       svfWveg, svfaveg, svfEaveg, svfSaveg, svfWaveg, svfNaveg, vegdem, vegdem2, albedo_b, absK, absL,
    #                       ewall, Fside, Fup, Fcyl, altitude, azimuth, zen, jday, usevegdem, onlyglobal, buildings, location, psi,
    #                       landcover, lc_grid, dectime, altmax, dirwalls, walls, cyl, elvis, Ta, RH, radG, radD, radI, P,
    #                       amaxvalue, bush, Twater, TgK, Tstart, alb_grid, emis_grid, TgK_wall, Tstart_wall, TmaxLST,
    #                       TmaxLST_wall, first, second, svfalfa, svfbuveg, firstdaytime, timeadd, timestepdec, Tgmap1,
    #                       Tgmap1E, Tgmap1S, Tgmap1W, Tgmap1N, CI, TgOut1, diffsh, ani):

    # This is the core function of the SOLWEIG model
    # 2016-Aug-28
    # Fredrik Lindberg, fredrikl@gvc.gu.se
    # Goteborg Urban Climate Group
    # Gothenburg University
    #
    # Input variables:
    # dsm = digital surface model
    # scale = height to pixel size (2m pixel gives scale = 0.5)
    # svf,svfN,svfW,svfE,svfS = SVFs for building and ground
    # svfveg,svfNveg,svfEveg,svfSveg,svfWveg = Veg SVFs blocking sky
    # svfaveg,svfEaveg,svfSaveg,svfWaveg,svfNaveg = Veg SVFs blocking buildings
    # vegdem = Vegetation canopy DSM
    # vegdem2 = Vegetation trunk zone DSM
    # albedo_b = building wall albedo
    # absK = human absorption coefficient for shortwave radiation
    # absL = human absorption coefficient for longwave radiation
    # ewall = Emissivity of building walls
    # Fside = The angular factors between a person and the surrounding surfaces
    # Fup = The angular factors between a person and the surrounding surfaces
    # Fcyl = The angular factors between a culidric person and the surrounding surfaces
    # altitude = Sun altitude (degree)
    # azimuth = Sun azimuth (degree)
    # zen = Sun zenith angle (radians)
    # jday = day of year
    # usevegdem = use vegetation scheme
    # onlyglobal = calculate dir and diff from global shortwave (Reindl et al. 1990)
    # buildings = Boolena grid to identify building pixels
    # location = geographic location
    # height = height of measurements point (center of gravity of human)
    # psi = 1 - Transmissivity of shortwave through vegetation
    # landcover = use landcover scheme !!!NEW IN 2015a!!!
    # lc_grid = grid with landcoverclasses
    # lc_class = table with landcover properties
    # dectime = decimal time
    # altmax = maximum sun altitude
    # dirwalls = aspect of walls
    # walls = one pixel row outside building footprint. height of building walls
    # cyl = consider man as cylinder instead of cude
    # elvis = dummy
    # Ta = air temp
    # RH
    # radG = global radiation
    # radD = diffuse
    # radI = direct
    # P = pressure
    # amaxvalue = max height of buildings
    # bush = grid representing bushes
    # Twater = temperature of water (daily)
    # TgK, Tstart, TgK_wall, Tstart_wall, TmaxLST,TmaxLST_wall,
    # alb_grid, emis_grid = albedo and emmissivity on ground
    # first, second = conneted to old Ts model (source area based on Smidt et al.)
    # svfalfa = SVF recalculated to angle
    # svfbuveg = complete SVF
    # firstdaytime, timeadd, timestepdec, Tgmap1, Tgmap1E, Tgmap1S, Tgmap1W, Tgmap1N,
    # CI = Clearness index
    # TgOut1 = old Ts model
    # diffsh, ani = Used in anisotrpic models (Wallenberg et al. 2019, 2022)

    # # # Core program start # # #
    # Instrument offset in degrees
    t = 0.0

    # Stefan Bolzmans Constant
    SBC = 5.67051e-8

    # Degrees to radians
    deg2rad = np.pi / 180

    # Find sunrise decimal hour - new from 2014a
    _, _, _, SNUP = daylen(jday, location["latitude"])

    # Vapor pressure
    ea = 6.107 * 10 ** ((7.5 * Ta) / (237.3 + Ta)) * (RH / 100.0)

    # Determination of clear - sky emissivity from Prata (1996)
    msteg = 46.5 * (ea / (Ta + 273.15))
    esky = (1 - (1 + msteg) * np.exp(-((1.2 + 3.0 * msteg) ** 0.5))) + elvis  # -0.04 old error from Jonsson et al.2006

    if altitude > 0:  # # # # # # DAYTIME # # # # # #
        # Clearness Index on Earth's surface after Crawford and Dunchon (1999) with a correction
        #  factor for low sun elevations after Lindberg et al.(2008)
        I0, CI, Kt, I0et, CIuncorr = clearnessindex_2013b(zen, jday, Ta, RH / 100.0, radG, location, P)
        if (CI > 1) or (np.inf == CI):
            CI = 1

        # Estimation of radD and radI if not measured after Reindl et al.(1990)
        if onlyglobal == 1:
            I0, CI, Kt, I0et, CIuncorr = clearnessindex_2013b(zen, jday, Ta, RH / 100.0, radG, location, P)
            if (CI > 1) or (np.inf == CI):
                CI = 1

            radI, radD = diffusefraction(radG, altitude, Kt, Ta, RH)

        # Diffuse Radiation
        # Anisotropic Diffuse Radiation after Perez et al. 1993
        if anisotropic_sky == 1:
            patchchoice = 1
            zenDeg = zen * (180 / np.pi)
            # Relative luminance
            lv, pc_, pb_ = Perez_v3(zenDeg, azimuth, radD, radI, jday, patchchoice, patch_option)
            # Total relative luminance from sky, i.e. from each patch, into each cell
            aniLum = np.zeros((rows, cols), dtype=np.float32)
            for idx in range(lv.shape[0]):
                aniLum += diffsh[:, :, idx] * lv[idx, 2]

            dRad = aniLum * radD  # Total diffuse radiation from sky into each cell
        else:
            dRad = radD * svfbuveg
            patchchoice = 1
            lv = None

        # Shadow  images
        if usevegdem == 1:
            vegsh, sh, _, wallsh, wallsun, wallshve, _, facesun, wallsh_ = shadowingfunction_wallheight_23(
                dsm,
                vegdem,
                vegdem2,
                azimuth,
                altitude,
                scale,
                amaxvalue,
                bush,
                walls,
                dirwalls * np.pi / 180.0,
                walls_scheme,
                dirwalls_scheme * np.pi / 180.0,
            )
            shadow = sh - (1 - vegsh) * (1 - psi)
        else:
            sh, wallsh, wallsun, facesh, facesun = shadowingfunction_wallheight_13(
                dsm, azimuth, altitude, scale, walls, dirwalls * np.pi / 180.0
            )
            shadow = sh

        # # # Surface temperature parameterisation during daytime # # # #
        # new using max sun alt.instead of  dfm
        # Tgamp = (TgK * altmax - Tstart) + Tstart # Old
        Tgamp = TgK * altmax + Tstart  # Fixed 2021
        # Tgampwall = (TgK_wall * altmax - (Tstart_wall)) + (Tstart_wall) # Old
        Tgampwall = TgK_wall * altmax + Tstart_wall
        Tg = Tgamp * np.sin(
            (((dectime - np.floor(dectime)) - SNUP / 24) / (TmaxLST / 24 - SNUP / 24)) * np.pi / 2
        )  # 2015 a, based on max sun altitude
        Tgwall = Tgampwall * np.sin(
            (((dectime - np.floor(dectime)) - SNUP / 24) / (TmaxLST_wall / 24 - SNUP / 24)) * np.pi / 2
        )  # 2015a, based on max sun altitude

        if Tgwall < 0:  # temporary for removing low Tg during morning 20130205
            # Tg = 0
            Tgwall = 0

        # New estimation of Tg reduction for non - clear situation based on Reindl et al.1990
        radI0, _ = diffusefraction(I0, altitude, 1.0, Ta, RH)
        corr = 0.1473 * np.log(90 - (zen / np.pi * 180)) + 0.3454  # 20070329 correction of lat, Lindberg et al. 2008
        CI_Tg = (radG / radI0) + (1 - corr)
        if (CI_Tg > 1) or (CI_Tg == np.inf):
            CI_Tg = 1

        radG0 = radI0 * (np.sin(altitude * deg2rad)) + _
        CI_TgG = (radG / radG0) + (1 - corr)
        if (CI_TgG > 1) or (CI_TgG == np.inf):
            CI_TgG = 1

        # Tg = Tg * CI_Tg  # new estimation
        # Tgwall = Tgwall * CI_Tg
        Tg = Tg * CI_TgG  # new estimation
        Tgwall = Tgwall * CI_TgG
        if landcover == 1:
            Tg[Tg < 0] = 0  # temporary for removing low Tg during morning 20130205

        # # # # Ground View Factors # # # #
        (
            gvfLup,
            gvfalb,
            gvfalbnosh,
            gvfLupE,
            gvfalbE,
            gvfalbnoshE,
            gvfLupS,
            gvfalbS,
            gvfalbnoshS,
            gvfLupW,
            gvfalbW,
            gvfalbnoshW,
            gvfLupN,
            gvfalbN,
            gvfalbnoshN,
            gvfSum,
            gvfNorm,
        ) = gvf_2018a(
            wallsun,
            walls,
            buildings,
            scale,
            shadow,
            first,
            second,
            dirwalls,
            Tg,
            Tgwall,
            Ta,
            emis_grid,
            ewall,
            alb_grid,
            SBC,
            albedo_b,
            rows,
            cols,
            Twater,
            lc_grid,
            landcover,
        )

        # # # # Lup, daytime # # # #
        # Surface temperature wave delay - new as from 2014a
        Lup, timeaddnotused, Tgmap1 = TsWaveDelay_2015a(gvfLup, firstdaytime, timeadd, timestepdec, Tgmap1)
        LupE, timeaddnotused, Tgmap1E = TsWaveDelay_2015a(gvfLupE, firstdaytime, timeadd, timestepdec, Tgmap1E)
        LupS, timeaddnotused, Tgmap1S = TsWaveDelay_2015a(gvfLupS, firstdaytime, timeadd, timestepdec, Tgmap1S)
        LupW, timeaddnotused, Tgmap1W = TsWaveDelay_2015a(gvfLupW, firstdaytime, timeadd, timestepdec, Tgmap1W)
        LupN, timeaddnotused, Tgmap1N = TsWaveDelay_2015a(gvfLupN, firstdaytime, timeadd, timestepdec, Tgmap1N)

        # # For Tg output in POIs
        TgTemp = Tg * shadow + Ta
        TgOut, timeadd, TgOut1 = TsWaveDelay_2015a(
            TgTemp, firstdaytime, timeadd, timestepdec, TgOut1
        )  # timeadd only here v2021a

        # Building height angle from svf
        F_sh = cylindric_wedge(zen, svfalfa, rows, cols)  # Fraction shadow on building walls based on sun alt and svf
        F_sh[np.isnan(F_sh)] = 0.5

        # # # # # # # Calculation of shortwave daytime radiative fluxes # # # # # # #
        Kdown = (
            radI * shadow * np.sin(altitude * (np.pi / 180))
            + dRad
            + albedo_b * (1 - svfbuveg) * (radG * (1 - F_sh) + radD * F_sh)
        )  # *sin(altitude(i) * (pi / 180))

        Kup, KupE, KupS, KupW, KupN = Kup_veg_2015a(
            radI,
            radD,
            radG,
            altitude,
            svfbuveg,
            albedo_b,
            F_sh,
            gvfalb,
            gvfalbE,
            gvfalbS,
            gvfalbW,
            gvfalbN,
            gvfalbnosh,
            gvfalbnoshE,
            gvfalbnoshS,
            gvfalbnoshW,
            gvfalbnoshN,
        )

        Keast, Ksouth, Kwest, Knorth, KsideI, KsideD, Kside = Kside_veg_v2022a(
            radI,
            radD,
            radG,
            shadow,
            svfS,
            svfW,
            svfN,
            svfE,
            svfEveg,
            svfSveg,
            svfWveg,
            svfNveg,
            azimuth,
            altitude,
            psi,
            t,
            albedo_b,
            F_sh,
            KupE,
            KupS,
            KupW,
            KupN,
            cyl,
            lv,
            anisotropic_sky,
            diffsh,
            rows,
            cols,
            asvf,
            shmat,
            vegshmat,
            vbshvegshmat,
        )

        firstdaytime = 0

    else:  # # # # # # # NIGHTTIME # # # # # # # #
        Tgwall = 0
        # CI_Tg = -999  # F_sh = []

        # Nocturnal K fluxes set to 0
        Knight = np.zeros((rows, cols), dtype=np.float32)
        Kdown = np.zeros((rows, cols), dtype=np.float32)
        Kwest = np.zeros((rows, cols), dtype=np.float32)
        Kup = np.zeros((rows, cols), dtype=np.float32)
        Keast = np.zeros((rows, cols), dtype=np.float32)
        Ksouth = np.zeros((rows, cols), dtype=np.float32)
        Knorth = np.zeros((rows, cols), dtype=np.float32)
        KsideI = np.zeros((rows, cols), dtype=np.float32)
        KsideD = np.zeros((rows, cols), dtype=np.float32)
        F_sh = np.zeros((rows, cols), dtype=np.float32)
        Tg = np.zeros((rows, cols), dtype=np.float32)
        shadow = np.zeros((rows, cols), dtype=np.float32)
        CI_Tg = deepcopy(CI)
        CI_TgG = deepcopy(CI)
        dRad = np.zeros((rows, cols), dtype=np.float32)
        Kside = np.zeros((rows, cols), dtype=np.float32)

        # # # # Lup # # # #
        Lup = SBC * emis_grid * ((Knight + Ta + Tg + 273.15) ** 4)
        if landcover == 1:
            Lup[lc_grid == 3] = SBC * 0.98 * (Twater + 273.15) ** 4  # nocturnal Water temp

        LupE = Lup
        LupS = Lup
        LupW = Lup
        LupN = Lup

        # # For Tg output in POIs
        TgOut = Ta + Tg

        I0 = 0
        timeadd = 0
        firstdaytime = 1

    # # # # Ldown # # # #
    Ldown = (
        (svf + svfveg - 1) * esky * SBC * ((Ta + 273.15) ** 4)
        + (2 - svfveg - svfaveg) * ewall * SBC * ((Ta + 273.15) ** 4)
        + (svfaveg - svf) * ewall * SBC * ((Ta + 273.15 + Tgwall) ** 4)
        + (2 - svf - svfveg) * (1 - ewall) * esky * SBC * ((Ta + 273.15) ** 4)
    )  # Jonsson et al.(2006)
    # Ldown = Ldown - 25 # Shown by Jonsson et al.(2006) and Duarte et al.(2006)

    if CI < 0.95:  # non - clear conditions
        c = 1 - CI
        Ldown = Ldown * (1 - c) + c * (
            (svf + svfveg - 1) * SBC * ((Ta + 273.15) ** 4)
            + (2 - svfveg - svfaveg) * ewall * SBC * ((Ta + 273.15) ** 4)
            + (svfaveg - svf) * ewall * SBC * ((Ta + 273.15 + Tgwall) ** 4)
            + (2 - svf - svfveg) * (1 - ewall) * SBC * ((Ta + 273.15) ** 4)
        )  # NOT REALLY TESTED!!! BUT MORE CORRECT?

    # # # # Lside # # # #
    Least, Lsouth, Lwest, Lnorth = Lside_veg_v2022a(
        svfS,
        svfW,
        svfN,
        svfE,
        svfEveg,
        svfSveg,
        svfWveg,
        svfNveg,
        svfEaveg,
        svfSaveg,
        svfWaveg,
        svfNaveg,
        azimuth,
        altitude,
        Ta,
        Tgwall,
        SBC,
        ewall,
        Ldown,
        esky,
        t,
        F_sh,
        CI,
        LupE,
        LupS,
        LupW,
        LupN,
        anisotropic_sky,
    )

    # New parameterization scheme for wall temperatures
    if wallScheme == 1:
        # albedo_g = 0.15 #TODO Change to correct
        if altitude < 0:
            wallsh_ = 0
        voxelTable = wall_surface_temperature(
            voxelTable, wallsh_, altitude, azimuth, timeStep, radI, radD, radG, Ldown, Lup, Ta, esky
        )
    # Anisotropic sky
    if anisotropic_sky == 1:
        if "lv" not in locals():
            # Creating skyvault of patches of constant radians (Tregeneza and Sharples, 1993)
            skyvaultalt, skyvaultazi, _, _, _, _, _ = create_patches(patch_option)

            patch_emissivities = np.zeros(skyvaultalt.shape[0], dtype=np.float32)

            x = np.transpose(np.atleast_2d(skyvaultalt))
            y = np.transpose(np.atleast_2d(skyvaultazi))
            z = np.transpose(np.atleast_2d(patch_emissivities))

            L_patches = np.append(np.append(x, y, axis=1), z, axis=1)

        else:
            L_patches = deepcopy(lv)

        # Calculate steradians for patches if it is the first model iteration
        if i == 0:
            steradians, skyalt, patch_altitude = patch_steradians(L_patches)

        # Create lv from L_patches if nighttime, i.e. lv does not exist
        if altitude < 0:
            # CI = deepcopy(CI)
            lv = deepcopy(L_patches)
            KupE = 0
            KupS = 0
            KupW = 0
            KupN = 0

        # Adjust sky emissivity under semi-cloudy/hazy/cloudy/overcast conditions, i.e. CI lower than 0.95
        if CI < 0.95:
            esky_c = CI * esky + (1 - CI) * 1.0
            esky = esky_c

        (
            Ldown,
            Lside,
            Lside_sky,
            Lside_veg,
            Lside_sh,
            Lside_sun,
            Lside_ref,
            Least_,
            Lwest_,
            Lnorth_,
            Lsouth_,
            Keast,
            Ksouth,
            Kwest,
            Knorth,
            KsideI,
            KsideD,
            Kside,
            steradians,
            skyalt,
        ) = ani_sky(
            shmat,
            vegshmat,
            vbshvegshmat,
            altitude,
            azimuth,
            asvf,
            cyl,
            esky,
            L_patches,
            wallScheme,
            voxelTable,
            voxelMaps,
            steradians,
            Ta,
            Tgwall,
            ewall,
            Lup,
            radI,
            radD,
            radG,
            lv,
            albedo_b,
            0,
            diffsh,
            shadow,
            KupE,
            KupS,
            KupW,
            KupN,
            i,
        )
    else:
        Lside = np.zeros((rows, cols), dtype=np.float32)
        L_patches = None

    # Box and anisotropic longwave
    if cyl == 0 and anisotropic_sky == 1:
        Least += Least_
        Lwest += Lwest_
        Lnorth += Lnorth_
        Lsouth += Lsouth_

    # # # # Calculation of radiant flux density and Tmrt # # # #
    # Human body considered as a cylinder with isotropic all-sky diffuse
    if cyl == 1 and anisotropic_sky == 0:
        Sstr = absK * (KsideI * Fcyl + (Kdown + Kup) * Fup + (Knorth + Keast + Ksouth + Kwest) * Fside) + absL * (
            (Ldown + Lup) * Fup + (Lnorth + Least + Lsouth + Lwest) * Fside
        )
    # Human body considered as a cylinder with Perez et al. (1993) (anisotropic sky diffuse)
    # and Martin and Berdahl (1984) (anisotropic sky longwave)
    elif cyl == 1 and anisotropic_sky == 1:
        Sstr = absK * (Kside * Fcyl + (Kdown + Kup) * Fup + (Knorth + Keast + Ksouth + Kwest) * Fside) + absL * (
            (Ldown + Lup) * Fup + Lside * Fcyl + (Lnorth + Least + Lsouth + Lwest) * Fside
        )
    # Knorth = nan Ksouth = nan Kwest = nan Keast = nan
    else:  # Human body considered as a standing cube
        Sstr = absK * ((Kdown + Kup) * Fup + (Knorth + Keast + Ksouth + Kwest) * Fside) + absL * (
            (Ldown + Lup) * Fup + (Lnorth + Least + Lsouth + Lwest) * Fside
        )

    Tmrt = np.sqrt(np.sqrt(Sstr / (absL * SBC))) - 273.2

    # Add longwave to cardinal directions for output in POI
    if (cyl == 1) and (anisotropic_sky == 1):
        Least += Least_
        Lwest += Lwest_
        Lnorth += Lnorth_
        Lsouth += Lsouth_

    return (
        Tmrt,
        Kdown,
        Kup,
        Ldown,
        Lup,
        Tg,
        ea,
        esky,
        I0,
        CI,
        shadow,
        firstdaytime,
        timestepdec,
        timeadd,
        Tgmap1,
        Tgmap1E,
        Tgmap1S,
        Tgmap1W,
        Tgmap1N,
        Keast,
        Ksouth,
        Kwest,
        Knorth,
        Least,
        Lsouth,
        Lwest,
        Lnorth,
        KsideI,
        TgOut1,
        TgOut,
        radI,
        radD,
        Lside,
        L_patches,
        CI_Tg,
        CI_TgG,
        KsideD,
        dRad,
        Kside,
        steradians,
        voxelTable,
    )



# Generated with SMOP  0.41
from libsmop import *
# CRT_Acyl.m

    
@function
def CRT_Acyl(L=None,D=None,*args,**kwargs):
    varargin = CRT_Acyl.varargin
    nargin = CRT_Acyl.nargin

    #Function written August 2009 by Jenni Vanos for calculating the Area of
#the CRT for use in the CNR Comfa model for finidng Rabs.
    
    #L = length (m) - #for CRT this is ~0.10m
#D = diamater (m) - #for CRT this is ~0.01m
    
    pi=3.14
# CRT_Acyl.m:10
    Acyl=multiply(multiply(2,pi),(D / 2) ** 2) + multiply(multiply(dot(2.0,pi),(D / 2)),L)
# CRT_Acyl.m:12


import numpy as np
from tqdm import tqdm

from umep.util import shadowingfunctions as shadow
from umep.util.SEBESOLWEIGCommonFiles.create_patches import create_patches


def annulus_weight(altitude, aziinterval):
    n = 90.0
    steprad = (360.0 / aziinterval) * (np.pi / 180.0)
    annulus = 91.0 - altitude
    w = (
        (1.0 / (2.0 * np.pi))
        * np.sin(np.pi / (2.0 * n))
        * np.sin((np.pi * (2.0 * annulus - 1.0)) / (2.0 * n))
    )
    weight = steprad * w

    return weight


def svfForProcessing153(dsm, vegdem, vegdem2, scale, usevegdem, amaxvalue):
    # memory
    dsm = dsm.astype(np.float32)
    vegdem = vegdem.astype(np.float32)
    vegdem2 = vegdem2.astype(np.float32)
    # setup
    rows = dsm.shape[0]
    cols = dsm.shape[1]
    svf = np.zeros([rows, cols], dtype=np.float32)
    svfE = np.zeros([rows, cols], dtype=np.float32)
    svfS = np.zeros([rows, cols], dtype=np.float32)
    svfW = np.zeros([rows, cols], dtype=np.float32)
    svfN = np.zeros([rows, cols], dtype=np.float32)
    svfveg = np.zeros((rows, cols), dtype=np.float32)
    svfEveg = np.zeros((rows, cols), dtype=np.float32)
    svfSveg = np.zeros((rows, cols), dtype=np.float32)
    svfWveg = np.zeros((rows, cols), dtype=np.float32)
    svfNveg = np.zeros((rows, cols), dtype=np.float32)
    svfaveg = np.zeros((rows, cols), dtype=np.float32)
    svfEaveg = np.zeros((rows, cols), dtype=np.float32)
    svfSaveg = np.zeros((rows, cols), dtype=np.float32)
    svfWaveg = np.zeros((rows, cols), dtype=np.float32)
    svfNaveg = np.zeros((rows, cols), dtype=np.float32)

    # raster preprocessing handled upstream - don't duplicate

    # % Bush separation
    bush = np.logical_not(vegdem2 * vegdem) * vegdem

    index = 0

    # patch_option = 1 # 145 patches
    patch_option = 2  # 153 patches
    # patch_option = 3 # 306 patches
    # patch_option = 4 # 612 patches

    # Create patches based on patch_option
    (
        skyvaultalt,
        skyvaultazi,
        annulino,
        skyvaultaltint,
        aziinterval,
        skyvaultaziint,
        azistart,
    ) = create_patches(patch_option)

    skyvaultaziint = np.array([360 / patches for patches in aziinterval])
    iazimuth = np.hstack(np.zeros((1, np.sum(aziinterval))))  # Nils

    # float 32 for memory
    shmat = np.zeros((rows, cols, np.sum(aziinterval)), dtype=np.float32)
    vegshmat = np.zeros((rows, cols, np.sum(aziinterval)), dtype=np.float32)
    vbshvegshmat = np.zeros((rows, cols, np.sum(aziinterval)), dtype=np.float32)

    for j in range(0, skyvaultaltint.shape[0]):
        for k in range(0, int(360 / skyvaultaziint[j])):
            iazimuth[index] = k * skyvaultaziint[j] + azistart[j]
            if iazimuth[index] > 360.0:
                iazimuth[index] = iazimuth[index] - 360.0
            index = index + 1

    # NOTE: total for progress
    total = 0
    for i in range(0, skyvaultaltint.shape[0]):
        for j in np.arange(0, aziinterval[int(i)]):
            total += 1
    progress = tqdm(total=total)
    #
    aziintervalaniso = np.ceil(aziinterval / 2.0)
    index = int(0)
    for i in range(0, skyvaultaltint.shape[0]):
        for j in np.arange(0, aziinterval[int(i)]):
            altitude = skyvaultaltint[int(i)]
            azimuth = iazimuth[int(index)]

            # Casting shadow
            if usevegdem == 1:
                # numba doesn't seem to offer notable gains in this instance
                shadowresult = shadow.shadowingfunction_20(
                    dsm,
                    vegdem,
                    vegdem2,
                    azimuth,
                    altitude,
                    scale,
                    amaxvalue,
                    bush,
                    1,  # for svf
                )
                vegsh = shadowresult["vegsh"]
                vbshvegsh = shadowresult["vbshvegsh"]
                sh = shadowresult["sh"]
                vegshmat[:, :, index] = vegsh
                vbshvegshmat[:, :, index] = vbshvegsh
            else:
                sh = shadow.shadowingfunctionglobalradiation(
                    dsm, azimuth, altitude, scale, 1
                )
            shmat[:, :, index] = sh

            # Calculate svfs
            for k in np.arange(annulino[int(i)] + 1, (annulino[int(i + 1.0)]) + 1):
                weight = annulus_weight(k, aziinterval[i]) * sh
                svf = svf + weight
                weight = annulus_weight(k, aziintervalaniso[i]) * sh
                if (azimuth >= 0) and (azimuth < 180):
                    svfE = svfE + weight
                if (azimuth >= 90) and (azimuth < 270):
                    svfS = svfS + weight
                if (azimuth >= 180) and (azimuth < 360):
                    svfW = svfW + weight
                if (azimuth >= 270) or (azimuth < 90):
                    svfN = svfN + weight

            if usevegdem == 1:
                for k in np.arange(annulino[int(i)] + 1, (annulino[int(i + 1.0)]) + 1):
                    # % changed to include 90
                    weight = annulus_weight(k, aziinterval[i])
                    svfveg = svfveg + weight * vegsh
                    svfaveg = svfaveg + weight * vbshvegsh
                    weight = annulus_weight(k, aziintervalaniso[i])
                    if (azimuth >= 0) and (azimuth < 180):
                        svfEveg = svfEveg + weight * vegsh
                        svfEaveg = svfEaveg + weight * vbshvegsh
                    if (azimuth >= 90) and (azimuth < 270):
                        svfSveg = svfSveg + weight * vegsh
                        svfSaveg = svfSaveg + weight * vbshvegsh
                    if (azimuth >= 180) and (azimuth < 360):
                        svfWveg = svfWveg + weight * vegsh
                        svfWaveg = svfWaveg + weight * vbshvegsh
                    if (azimuth >= 270) or (azimuth < 90):
                        svfNveg = svfNveg + weight * vegsh
                        svfNaveg = svfNaveg + weight * vbshvegsh

            index += 1

            # track progress
            progress.update(1)

    svfS = svfS + 3.0459e-004
    svfW = svfW + 3.0459e-004
    # % Last azimuth is 90. Hence, manual add of last annuli for svfS and SVFW
    # %Forcing svf not be greater than 1 (some MATLAB crazyness)
    svf[(svf > 1.0)] = 1.0
    svfE[(svfE > 1.0)] = 1.0
    svfS[(svfS > 1.0)] = 1.0
    svfW[(svfW > 1.0)] = 1.0
    svfN[(svfN > 1.0)] = 1.0

    if usevegdem == 1:
        last = np.zeros((rows, cols))
        last[(vegdem2 == 0.0)] = 3.0459e-004
        svfSveg = svfSveg + last
        svfWveg = svfWveg + last
        svfSaveg = svfSaveg + last
        svfWaveg = svfWaveg + last
        # %Forcing svf not be greater than 1 (some MATLAB crazyness)
        svfveg[(svfveg > 1.0)] = 1.0
        svfEveg[(svfEveg > 1.0)] = 1.0
        svfSveg[(svfSveg > 1.0)] = 1.0
        svfWveg[(svfWveg > 1.0)] = 1.0
        svfNveg[(svfNveg > 1.0)] = 1.0
        svfaveg[(svfaveg > 1.0)] = 1.0
        svfEaveg[(svfEaveg > 1.0)] = 1.0
        svfSaveg[(svfSaveg > 1.0)] = 1.0
        svfWaveg[(svfWaveg > 1.0)] = 1.0
        svfNaveg[(svfNaveg > 1.0)] = 1.0

    svfresult = {
        "svf": svf,
        "svfE": svfE,
        "svfS": svfS,
        "svfW": svfW,
        "svfN": svfN,
        "svfveg": svfveg,
        "svfEveg": svfEveg,
        "svfSveg": svfSveg,
        "svfWveg": svfWveg,
        "svfNveg": svfNveg,
        "svfaveg": svfaveg,
        "svfEaveg": svfEaveg,
        "svfSaveg": svfSaveg,
        "svfWaveg": svfWaveg,
        "svfNaveg": svfNaveg,
        "shmat": shmat,
        "vegshmat": vegshmat,
        "vbshvegshmat": vbshvegshmat,
    }

    return svfresult



# Generated with SMOP  0.41
from libsmop import *
# CNR_Kup.m

    
@function
def CNR_Kup(alpha=None,Kup=None,L=None,D=None,*args,**kwargs):
    varargin = CNR_Kup.varargin
    nargin = CNR_Kup.nargin

    # Used to calculate the total reflected solar radiation (K) absorbed by the
# a cylinder with inputs alpha (cylinder albedo), Kup (measured reflected solar
# radiation by the CNR net radiometer (W/m2), Acyl (area of the cylinder,
# m2)
    
    Acyl=CRT_Acyl(L,D)
# CNR_Kup.m:8
    
    Kup_abs=multiply(multiply(multiply((1 - alpha),Kup),0.5),Acyl)
# CNR_Kup.m:10
    
    # Written August, 2006 by Natasha Kenny
# Updated Comments Jan 2016 Jenni Vanos


import numpy as np
import matplotlib.cm as cmx
from matplotlib.pyplot import figure, show
import matplotlib.pyplot as plt
from matplotlib.patches import Circle
import matplotlib
import matplotlib.colors as colors

#def PolarBarPlot(lv, radD, outfolder, YYYY, DOY, XH, hours, XM, minu, iStep, idxStep):
def PolarBarPlot(lv, solar_altitude, solar_azimuth, fig_title, filename_out, minrad, maxrad, plot_type):
#def PolarBarPlot(lv, filename_start, outfolder, YYYY, DOY, XH, hours, XM, minu, iStep, minrad, maxrad):

    deg2rad = np.pi/180

    fig = figure()
    ax = fig.add_subplot(111, projection = 'polar')

    # Set zero location and clockwise
    ax.set_theta_zero_location("N")
    if plot_type: 
        ax.set_theta_direction('clockwise')
    else:
        ax.set_theta_direction('anticlockwise')
        # ax.set_theta_direction('clockwise')
    skyalt, skyalt_c = np.unique(lv[:, 0], return_counts=True)   # Unique altitudes in lv, i.e. unique altitude for the patches

    # lvSum = np.sum(lv[:,2])

    lv_norm = lv[:, 2]  # Watts per square meter Steradian
    lv_alt = lv[:, 0]
    # lv_azi = lv[:, 1]

    # lvMax = np.around(np.max(lv_norm), decimals=3)

    if plot_type:
        jet = cm = plt.get_cmap('jet')
        cNorm = colors.Normalize(vmin=minrad, vmax=maxrad) # Watts per square meter Steradian
        scalarMap = cmx.ScalarMappable(norm=cNorm, cmap=jet)
    else:
        # patch_category = {1.8:'deepskyblue', 2.5:'forestgreen', 3.3:'yellow', 4.5:'peru', 6.0:'grey'}
        patch_category = {1.8:'deepskyblue', 2.5:'forestgreen', 3.3:'yellow', 4.5:'peru'}

    if lv.shape[0] < 160:
        azistart = np.array([0, 4, 2, 5, 8, 0, 10, 0]) # Fredrik/Nils
    else:
        azistart = np.array([0, 0, 4, 4, 2, 2, 5, 5, 8, 8, 0, 0, 10, 10, 0]) # Nils test
    radii = np.ones(np.max(skyalt_c))*1.5
    radii_sub = 0
    for i in range(skyalt_c.__len__()):
        clrs = lv_norm[lv_alt == skyalt[i]]
        # print(clrs)
        if skyalt_c[i] > 1:
            theta_patch = ( np.arange(0, skyalt_c[skyalt == skyalt[i]][0], 1) * (360 / skyalt_c[skyalt == skyalt[i]][0] )) * deg2rad
            width_patch = ( np.ones( skyalt_c[skyalt == skyalt[i]][0] ) * ( 360 * deg2rad ) / skyalt_c[skyalt == skyalt[i]][0] )

            # if plot_type:
            #     patch_order_range = range(skyalt_c[skyalt == skyalt[i]][0])
            # else:
            #     patch_order_range = reversed(range(skyalt_c[skyalt == skyalt[i]][0]))

            patch_order_range = range(skyalt_c[skyalt == skyalt[i]][0])

            for j in patch_order_range:
            # for j in range(skyalt_c[skyalt == skyalt[i]][0]):
                if plot_type:
                    patch_color = scalarMap.to_rgba(clrs[j])
                    # bars = ax.bar(theta_patch[j] + (azistart[i] * deg2rad), radii[j]-radii_sub, width=width_patch[j], bottom=0.0, edgecolor='black', facecolor=patch_color)
                else:
                    patch_color = patch_category[clrs[j]]
                    # bars = ax.bar((np.pi*2 - theta_patch[j]) + (azistart[i] * deg2rad), radii[j]-radii_sub, width=width_patch[j], bottom=0.0, edgecolor='black', facecolor=patch_color)
                # bars = ax.bar(theta_patch[j] + (azistart[i] * deg2rad) - width_patch[j], radii[j]-radii_sub, width=width_patch[j], bottom=0.0, edgecolor='black', facecolor=patch_color)
                bars = ax.bar(theta_patch[j] + (azistart[i] * deg2rad), radii[j]-radii_sub, width=width_patch[j], bottom=0.0, edgecolor='black', facecolor=patch_color)
            if lv.shape[0] < 160:
                radii_sub += 0.2
            else:
                radii_sub += 0.1 # Nils test
        else:
            # Create a circle for the center of the plot
            my_circle=Circle( (0,0), 0.1, edgecolor='black', transform=ax.transData._b)
            if plot_type:
                patch_color = scalarMap.to_rgba(clrs[0])
            else:
                patch_color = patch_category[clrs[0]]
            my_circle.set_facecolor(patch_color)
            ax.add_artist(my_circle)

    # Adding sun's position
    if ((plot_type == 0) & (solar_altitude > 0)):
        sun_position = plt.scatter((solar_azimuth * deg2rad), 1.5 - (1.5 * (solar_altitude/90)), s=(70 * (solar_altitude/90)), color='gold')
        sun_position_star1 = plt.scatter((solar_azimuth * deg2rad), 1.5 - (1.5 * (solar_altitude/90)), s=(400 * (solar_altitude/90)), color='gold', marker='1')
        sun_position_star2 = plt.scatter((solar_azimuth * deg2rad), 1.5 - (1.5 * (solar_altitude/90)), s=(400 * (solar_altitude/90)), color='gold', marker='2')
        # sun_position = plt.scatter(np.pi * 2 - (solar_azimuth * deg2rad), 1.5 - (1.5 * (solar_altitude/90)), s=(70 * (solar_altitude/90)), color='gold')
        # sun_position_star1 = plt.scatter(np.pi * 2 - (solar_azimuth * deg2rad), 1.5 - (1.5 * (solar_altitude/90)), s=(400 * (solar_altitude/90)), color='gold', marker='1')
        # sun_position_star2 = plt.scatter(np.pi * 2 - (solar_azimuth * deg2rad), 1.5 - (1.5 * (solar_altitude/90)), s=(400 * (solar_altitude/90)), color='gold', marker='2')
        ax.add_artist(sun_position_star1)
        ax.add_artist(sun_position_star2)
        ax.add_artist(sun_position)

    # Remove grids and tick labels
    ax.set_yticklabels([])
    ax.grid(False)
    ax.spines['polar'].set_visible(False)

    if plot_type:
        # Adding colorbar
        sm = plt.cm.ScalarMappable(cmap=plt.cm.jet, norm=plt.Normalize(vmin=minrad, vmax=maxrad)) # Watts per square meter Steradian
        sm._A = []
        cbaxes = fig.add_axes([0.87, 0.1, 0.03, 0.8])
        cb = plt.colorbar(sm, cax = cbaxes)
        cb.ax.set_title(r'$W/m^2$ $sr^{-1}$', fontsize=12, fontweight='bold') # Watts per square meter Steradian
    else:
        # legend_colors = {'Shaded building wall':'peru', 'Sunlit building wall':'yellow', 'Sky':'deepskyblue', 'Trees':'forestgreen', 'Roof':'black'}
        # legend_colors = {'Building wall':'peru', 'Building roof':'grey', 'Trees':'forestgreen', 'Sky':'deepskyblue'}
        legend_colors = {'Building wall':'peru', 'Trees':'forestgreen', 'Sky':'deepskyblue'}
        legend_labels = list(legend_colors.keys())
        legend_handles = [plt.Rectangle((0,0),1,1, color=legend_colors[legend_label]) for legend_label in legend_labels]
        if solar_altitude > 0:
            legend_handles.extend([(sun_position, sun_position_star1, sun_position_star2)])
            legend_labels.extend(['Sun'])
        plt.legend(legend_handles, legend_labels,
                    title='Categories',
                    loc='lower left',
                    bbox_to_anchor=(0.95, 0, 0.5, 1),
                    fontsize=7,
                    frameon=False,
                    title_fontsize=7,
                    markerscale=0.7)

    ax.set_title(fig_title, weight='bold', fontsize=10)

    fig.savefig(filename_out, dpi=300)
    plt.close('all')


from __future__ import division
import numpy as np
from ...util.SEBESOLWEIGCommonFiles.create_patches import create_patches

def Perez_v3(zen, azimuth, radD, radI, jday, patchchoice, patch_option):
    """
    This function calculates distribution of luminance on the skyvault based on
    Perez luminince distribution model.
    
    ALL-WEATHER MODEL FOR SKY LUMINANCE DISTRIBUTION
    PRELIMINARY CONFIGURATION AND VALIDATION
    R. PEREZ, R. SEALS, and J. MICHALSKY
    Solar Energy Vol. 50, No, 3, pp. 235-245, 1993 
    
    Abstract--This article reports the development and evaluation of a new model for describing, from routine
    irradiance measurements, the mean instantaneous sky luminance angular distribution patterns for all sky
    conditions from overcast to clear, through partly cloudy, skies.

    Created by:
    Fredrik Lindberg 20120527, fredrikl@gvc.gu.se
    Gothenburg University, Sweden
    Urban Climte Group
    
    Input parameters:
     - zen:     Zenith angle of the Sun (in degrees)
     - azimuth: Azimuth angle of the Sun (in degrees)
     - radD:    Horizontal diffuse radiation (W m-2)
     - radI:    Direct radiation perpendicular to the Sun beam (W m-2)
     - jday:    Day of year
    
    Output parameters:
     - lv:   Relative luminance map (same dimensions as theta. gamma)
    

    acoeff=[1.353 -0.258 -0.269 -1.437
           -1.222 -0.773 1.415 1.102
           -1.100 -0.252 0.895 0.016
           -0.585 -0.665 -0.267 0.712
           -0.600 -0.347 -2.500 2.323
           -1.016 -0.367 1.008 1.405
           -1.000 0.021 0.503 -0.512
           -1.050 0.029 0.426 0.359];
    
    bcoeff=[-0.767 0.001 1.273 -0.123
            -0.205 0.037 -3.913 0.916
             0.278 -0.181 -4.500 1.177
             0.723 -0.622 -5.681 2.630
             0.294 0.049 -5.681 1.842
             0.288 -0.533 -3.850 3.375
            -0.300 0.192 0.702 -1.632
            -0.325 0.116 0.778 0.003];
    
    ccoeff=[2.800 0.600 1.238 1.000
            6.975 0.177 6.448 -0.124
            24.22 -13.08 -37.70 34.84
            33.34 -18.30 -62.25 52.08
            21.00 -4.766 -21.59 7.249
            14.00 -0.999 -7.14 7.547
            19.00 -5.000 1.243 -1.91
            31.06 -14.50 -46.11 55.37];
    
    dcoeff=[1.874 0.630 0.974 0.281
           -1.580 -0.508 -1.781 0.108
           -5.00 1.522 3.923 -2.62
           -3.50 0.002 1.148 0.106
           -3.50 -0.155 1.406 0.399
           -3.40 -0.108 -1.075 1.57
           -4.00 0.025 0.384 0.266
           -7.23 0.405 13.35 0.623];
    
    ecoeff=[0.035 -0.125 -0.572 0.994
            0.262 0.067 -0.219 -0.428
           -0.016 0.160 0.420 -0.556
            0.466 -0.33 -0.088 -0.033
            0.003 0.077 -0.066 -0.129
           -0.067 0.402 0.302 -0.484
            1.047 -0.379 -2.452 1.466
            1.500 -0.643 1.856 0.564];

    :param zen:
    :param azimuth:
    :param radD:
    :param radI:
    :param jday:
    :param patchchoice:
    :return:
    """

    m_a1 = np.array([1.3525, -1.2219, -1.1000, -0.5484, -0.6000, -1.0156, -1.0000, -1.0500])
    m_a2 = np.array([-0.2576, -0.7730, -0.2515, -0.6654, -0.3566, -0.3670, 0.0211, 0.0289])
    m_a3 = np.array([-0.2690, 1.4148, 0.8952, -0.2672, -2.5000, 1.0078, 0.5025, 0.4260])
    m_a4 = np.array([-1.4366, 1.1016, 0.0156, 0.7117, 2.3250, 1.4051, -0.5119, 0.3590])
    m_b1 = np.array([-0.7670, -0.2054, 0.2782, 0.7234, 0.2937, 0.2875, -0.3000, -0.3250])
    m_b2 = np.array([0.0007, 0.0367, -0.1812, -0.6219, 0.0496, -0.5328, 0.1922, 0.1156])
    m_b3 = np.array([1.2734, -3.9128, -4.5000, -5.6812, -5.6812, -3.8500, 0.7023, 0.7781])
    m_b4 = np.array([-0.1233, 0.9156, 1.1766, 2.6297, 1.8415, 3.3750, -1.6317, 0.0025])
    m_c1 = np.array([2.8000, 6.9750, 24.7219, 33.3389, 21.0000, 14.0000, 19.0000, 31.0625])
    m_c2 = np.array([0.6004, 0.1774, -13.0812, -18.3000, -4.7656, -0.9999, -5.0000, -14.5000])
    m_c3 = np.array([1.2375, 6.4477, -37.7000, -62.2500, -21.5906, -7.1406, 1.2438, -46.1148])
    m_c4 = np.array([1.0000, -0.1239, 34.8438, 52.0781, 7.2492, 7.5469, -1.9094, 55.3750])
    m_d1 = np.array([1.8734, -1.5798, -5.0000, -3.5000, -3.5000, -3.4000, -4.0000, -7.2312])
    m_d2 = np.array([0.6297, -0.5081, 1.5218, 0.0016, -0.1554, -0.1078, 0.0250, 0.4050])
    m_d3 = np.array([0.9738, -1.7812, 3.9229, 1.1477, 1.4062, -1.0750, 0.3844, 13.3500])
    m_d4 = np.array([0.2809, 0.1080, -2.6204, 0.1062, 0.3988, 1.5702, 0.2656, 0.6234])
    m_e1 = np.array([0.0356, 0.2624, -0.0156, 0.4659, 0.0032, -0.0672, 1.0468, 1.5000])
    m_e2 = np.array([-0.1246, 0.0672, 0.1597, -0.3296, 0.0766, 0.4016, -0.3788, -0.6426])
    m_e3 = np.array([-0.5718, -0.2190, 0.4199, -0.0876, -0.0656, 0.3017, -2.4517, 1.8564])
    m_e4 = np.array([0.9938, -0.4285, -0.5562, -0.0329, -0.1294, -0.4844, 1.4656, 0.5636])
    
    acoeff = np.transpose(np.atleast_2d([m_a1, m_a2, m_a3, m_a4]))
    bcoeff = np.transpose(np.atleast_2d([m_b1, m_b2, m_b3, m_b4]))
    ccoeff = np.transpose(np.atleast_2d([m_c1, m_c2, m_c3, m_c4]))
    dcoeff = np.transpose(np.atleast_2d([m_d1, m_d2, m_d3, m_d4]))
    ecoeff = np.transpose(np.atleast_2d([m_e1, m_e2, m_e3, m_e4]))

    deg2rad = np.pi/180
    rad2deg = 180/np.pi
    altitude = 90-zen
    zen = zen * deg2rad
    azimuth = azimuth * deg2rad
    altitude = altitude * deg2rad
    Idh = radD
    # Ibh = radI/sin(altitude)
    Ibn = radI

    # Skyclearness
    PerezClearness = ((Idh + Ibn) / Idh + 1.041 * np.power(zen, 3)) / (1 + 1.041 * np.power(zen, 3))
    # Extra terrestrial radiation
    day_angle = jday*2*np.pi/365
    #I0=1367*(1+0.033*np.cos((2*np.pi*jday)/365))
    I0 = 1367*(1.00011+0.034221*np.cos(day_angle) + 0.00128*np.sin(day_angle)+0.000719 *
               np.cos(2*day_angle)+0.000077*np.sin(2*day_angle))    # New from robinsson

    # Optical air mass
    # m=1/altitude; old
    if altitude >= 10*deg2rad:
        AirMass = 1/np.sin(altitude)
    elif altitude < 0:   # below equation becomes complex
        AirMass = 1/np.sin(altitude)+0.50572*np.power(180*complex(altitude)/np.pi+6.07995, -1.6364)
    else:
        # Added brackets to denominator: np.sin(0) was giving zero hence division by zero = infinity
        AirMass = 1 / (np.sin(altitude) + 0.50572 * np.power(180 * altitude / np.pi + 6.07995, -1.6364))

    # Skybrightness
    # if altitude*rad2deg+6.07995>=0
    PerezBrightness = (AirMass*radD)/I0
    if Idh <= 10:
        # m_a=0;m_b=0;m_c=0;m_d=0;m_e=0;
        PerezBrightness = 0
    #if altitude < 0:
        #print("Airmass")
        #print(AirMass)
        #print(PerezBrightness)
    # sky clearness bins
    if PerezClearness < 1.065:
        intClearness = 0
    elif PerezClearness < 1.230:
        intClearness = 1
    elif PerezClearness < 1.500:
        intClearness = 2
    elif PerezClearness < 1.950:
        intClearness = 3
    elif PerezClearness < 2.800:
        intClearness = 4
    elif PerezClearness < 4.500:
        intClearness = 5
    elif PerezClearness < 6.200:
        intClearness = 6
    if PerezClearness > 6.200:
        intClearness = 7

    m_a = acoeff[intClearness,  0] + acoeff[intClearness,  1] * zen + PerezBrightness * (acoeff[intClearness,  2] + acoeff[intClearness,  3] * zen)
    m_b = bcoeff[intClearness,  0] + bcoeff[intClearness,  1] * zen + PerezBrightness * (bcoeff[intClearness,  2] + bcoeff[intClearness,  3] * zen)
    m_e = ecoeff[intClearness,  0] + ecoeff[intClearness,  1] * zen + PerezBrightness * (ecoeff[intClearness,  2] + ecoeff[intClearness,  3] * zen)

    if intClearness > 0:
        m_c = ccoeff[intClearness, 0] + ccoeff[intClearness, 1] * zen + PerezBrightness * (ccoeff[intClearness, 2] + ccoeff[intClearness, 3] * zen)
        m_d = dcoeff[intClearness, 0] + dcoeff[intClearness, 1] * zen + PerezBrightness * (dcoeff[intClearness, 2] + dcoeff[intClearness, 3] * zen)
    else:
        # different equations for c & d in clearness bin no. 1,  from Robinsson
        m_c = np.exp(np.power(PerezBrightness * (ccoeff[intClearness, 0] + ccoeff[intClearness, 1] * zen), ccoeff[intClearness, 2]))-1
        m_d = -np.exp(PerezBrightness * (dcoeff[intClearness, 0] + dcoeff[intClearness, 1] * zen)) + dcoeff[intClearness, 2] + \
            PerezBrightness * dcoeff[intClearness, 3]

    # print 'a = ', m_a
    # print 'b = ', m_b
    # print 'e = ', m_e
    # print 'c = ', m_c
    # print 'd = ', m_d

    if patchchoice == 2:
        skyvaultalt = np.atleast_2d([])
        skyvaultazi = np.atleast_2d([])
        # Creating skyvault at one degree intervals
        skyvaultalt = np.ones([90, 361])*90
        skyvaultazi = np.empty((90, 361))
        for j in range(90):
            skyvaultalt[j, :] = 91-j
            skyvaultazi[j, :] = range(361)
            
    elif patchchoice == 1:
        # Creating skyvault of patches of constant radians (Tregeneza and Sharples, 1993)
        skyvaultalt, skyvaultazi, _, _, _, _, _ = create_patches(patch_option)

    skyvaultzen = (90 - skyvaultalt) * deg2rad
    skyvaultalt = skyvaultalt * deg2rad
    skyvaultazi = skyvaultazi * deg2rad

    # Angular distance from the sun from Robinsson
    cosSkySunAngle = np.sin(skyvaultalt) * np.sin(altitude) + \
                     np.cos(altitude) * np.cos(skyvaultalt) * np.cos(np.abs(skyvaultazi-azimuth))

    cos_zen = np.cos(skyvaultzen)
    exp_arg_h = m_b / cos_zen
    ang = np.arccos(cosSkySunAngle)
    exp_arg_c = m_d * ang
    circumsolar = 1 + m_c * np.exp(exp_arg_c) + m_e * cosSkySunAngle * cosSkySunAngle

    lv = (1 + m_a * np.exp(exp_arg_h)) * circumsolar

    # Normalisation (with safeguard)
    if np.any(lv < 0):
        print("Warning: found negative Perez luminances, using uniform distribution as fallback.")
        lv.fill(1.0 / lv.size)  # uniform fallback
    else:
        lv = lv / np.sum(lv)

    # plotting
    # axesm('stereo','Origin',[90 180],'MapLatLimit',[0 90],'Aspect','transverse')
    # framem off; gridm on; mlabel off; plabel off;axis on;
    # setm(gca,'MLabelParallel',-20)
    # geoshow(skyvaultalt*rad2deg,skyvaultazi*rad2deg,lv,'DisplayType','texture');
    # colorbar
    # set(gcf,'Color',[1 1 1])
    # pause(1)

    if patchchoice == 1:
        #x = np.atleast_2d([])
        #lv = np.transpose(np.append(np.append(np.append(x, skyvaultalt*rad2deg), skyvaultazi*rad2deg), lv))
        x = np.transpose(np.atleast_2d(skyvaultalt*rad2deg))
        y = np.transpose(np.atleast_2d(skyvaultazi*rad2deg))
        z = np.transpose(np.atleast_2d(lv))
        lv = np.append(np.append(x, y, axis=1), z, axis=1)
    return lv, PerezClearness, PerezBrightness


from __future__ import division
import numpy as np
# import matplotlib.pylab as plt

def shade_on_walls(azimuth, aspect, walls, dsm, f, shvoveg):
    # wall shadows wall parameterization
    wallbol = (walls > 0).astype(float)
    
    # Removing walls in shadow due to selfshadowing
    azilow = azimuth - np.pi/2
    azihigh = azimuth + np.pi/2
    if azilow >= 0 and azihigh < 2*np.pi:    # 90 to 270  (SHADOW)
        facesh = np.logical_or(aspect < azilow, aspect >= azihigh).astype(float) - wallbol + 1    # TODO check
    elif azilow < 0 and azihigh <= 2*np.pi:    # 0 to 90
        azilow = azilow + 2*np.pi
        facesh = np.logical_or(aspect > azilow, aspect <= azihigh) * -1 + 1    # (SHADOW)
    elif azilow > 0 and azihigh >= 2*np.pi:    # 270 to 360
        azihigh -= 2 * np.pi
        facesh = np.logical_or(aspect > azilow, aspect <= azihigh)*-1 + 1    # (SHADOW)

    shvo = f - dsm   # building shadow volume
    facesun = np.logical_and(facesh + (walls > 0).astype(float) == 1, walls > 0).astype(float)
    wallsun = np.copy(walls-shvo)
    wallsun[wallsun < 0] = 0
    wallsun[facesh == 1] = 0    # Removing walls in "self"-shadow
    wallsh = np.copy(walls-wallsun)

    wallshve = shvoveg * wallbol
    wallshve = wallshve - wallsh
    wallshve[wallshve < 0] = 0
    id = np.where(wallshve > walls)
    wallshve[id] = walls[id]
    wallsun = wallsun-wallshve    # problem with wallshve only
    id = np.where(wallsun < 0)
    wallshve[id] = 0
    wallsun[id] = 0
    # if np.sum(wallshve <= 0) == wallshve.size:
    #     wallshve[:, :] = 0    

    return wallsh, wallsun, wallshve, facesh, facesun

def shadowingfunction_wallheight_23(a, vegdem, vegdem2, azimuth, altitude, scale, amaxvalue, bush, walls, aspect, walls_scheme=False, aspect_scheme=False):
    """
    This function calculates shadows on a DSM and shadow height on building
    walls including both buildings and vegetion units.
    New functionallity to deal with pergolas, August 2021
    
    INPUTS:
    a = DSM
    vegdem = Vegetation canopy DSM (magl)
    vegdem2 = Trunkzone DSM (magl)
    azimuth and altitude = sun position
    scale= scale of DSM (1 meter pixels=1, 2 meter pixels=0.5)
    walls= pixel row 'outside' buildings. will be calculated if empty
    aspect=normal aspect of walls
    
    OUTPUT:
    sh=ground and roof shadow

    wallsh = height of wall that is in shadow
    wallsun = hieght of wall that is in sun

    original Matlab code:
    Fredrik Lindberg 2013-08-14
    fredrikl@gvc.gu.se

    :param a:
    :param vegdem:
    :param vegdem2:
    :param azimuth:
    :param altitude:
    :param scale:
    :param amaxvalue:
    :param bush:
    :param walls:
    :param aspect:
    :return:
    """

    # conversion
    degrees = np.pi/180.
    azimuth *= degrees
    altitude *= degrees
    
    # measure the size of the image
    sizex = np.shape(a)[0]
    sizey = np.shape(a)[1]
    
    # initialise parameters
    dx = 0
    dy = 0
    dz = 0
    temp = np.zeros((sizex, sizey))
    tempvegdem = np.zeros((sizex, sizey))
    tempvegdem2 = np.zeros((sizex, sizey))
    templastfabovea = np.zeros((sizex, sizey))
    templastgabovea = np.zeros((sizex, sizey))
    bushplant = bush > 1
    sh = np.zeros((sizex, sizey)) #shadows from buildings
    vbshvegsh = np.copy(sh) #vegetation blocking buildings
    vegsh = np.add(np.zeros((sizex, sizey)), bushplant, dtype=float) #vegetation shadow
    f = np.copy(a)
    shvoveg = np.copy(vegdem) # for vegetation shadowvolume
    # g = np.copy(sh)
    wallbol = (walls > 0).astype(float)

    # other loop parameters
    pibyfour = np.pi/4
    threetimespibyfour = 3*pibyfour
    fivetimespibyfour = 5*pibyfour
    seventimespibyfour = 7*pibyfour
    sinazimuth = np.sin(azimuth)
    cosazimuth = np.cos(azimuth)
    tanazimuth = np.tan(azimuth)
    signsinazimuth = np.sign(sinazimuth)
    signcosazimuth = np.sign(cosazimuth)
    dssin = np.abs(1/sinazimuth)
    dscos = np.abs(1/cosazimuth)
    tanaltitudebyscale = np.tan(altitude)/scale

    index = 0

    # new case with pergola (thin vertical layer of vegetation), August 2021
    dzprev = 0

    # main loop
    while (amaxvalue >= dz) and (np.abs(dx) < sizex) and (np.abs(dy) < sizey):
        if ((pibyfour <= azimuth) and (azimuth < threetimespibyfour)) or ((fivetimespibyfour <= azimuth) and (azimuth < seventimespibyfour)):
            dy = signsinazimuth * index
            dx = -1 * signcosazimuth * np.abs(np.round(index / tanazimuth))
            ds = dssin
        else:
            dy = signsinazimuth * np.abs(np.round(index * tanazimuth))
            dx = -1 * signcosazimuth * index
            ds = dscos

        # note: dx and dy represent absolute values while ds is an incremental value
        dz = (ds * index) * tanaltitudebyscale
        tempvegdem[0:sizex, 0:sizey] = 0
        tempvegdem2[0:sizex, 0:sizey] = 0
        temp[0:sizex, 0:sizey] = 0
        templastfabovea[0:sizex, 0:sizey] = 0.
        templastgabovea[0:sizex, 0:sizey] = 0.
        absdx = np.abs(dx)
        absdy = np.abs(dy)
        xc1 = int((dx+absdx)/2)
        xc2 = int(sizex+(dx-absdx)/2)
        yc1 = int((dy+absdy)/2)
        yc2 = int(sizey+(dy-absdy)/2)
        xp1 = -int((dx-absdx)/2)
        xp2 = int(sizex-(dx+absdx)/2)
        yp1 = -int((dy-absdy)/2)
        yp2 = int(sizey-(dy+absdy)/2)

        tempvegdem[xp1:xp2, yp1:yp2] = vegdem[xc1:xc2, yc1:yc2] - dz
        tempvegdem2[xp1:xp2, yp1:yp2] = vegdem2[xc1:xc2, yc1:yc2] - dz
        temp[xp1:xp2, yp1:yp2] = a[xc1:xc2, yc1:yc2]-dz

        f = np.fmax(f, temp) #Moving building shadow
        shvoveg = np.fmax(shvoveg, tempvegdem) # moving vegetation shadow volume
        
        sh[f > a] = 1
        sh[f <= a] = 0   
        fabovea = (tempvegdem > a).astype(int)   #vegdem above DEM
        gabovea = (tempvegdem2 > a).astype(int)   #vegdem2 above DEM
        
        #new pergola condition
        templastfabovea[xp1:xp2, yp1:yp2] = vegdem[xc1:xc2, yc1:yc2]-dzprev
        templastgabovea[xp1:xp2, yp1:yp2] = vegdem2[xc1:xc2, yc1:yc2]-dzprev
        lastfabovea = templastfabovea > a
        lastgabovea = templastgabovea > a
        dzprev = dz
        vegsh2 = np.add(np.add(np.add(fabovea, gabovea, dtype=float),lastfabovea, dtype=float),lastgabovea, dtype=float)
        vegsh2[vegsh2 == 4] = 0.
        # vegsh2[vegsh2 == 1] = 0. # This one is the ultimate question...
        vegsh2[vegsh2 > 0] = 1.

        vegsh = np.fmax(vegsh, vegsh2)
        vegsh[vegsh*sh > 0] = 0    
        vbshvegsh = np.copy(vegsh) + vbshvegsh # removing shadows 'behind' buildings
    
        index += 1

    sh = 1-sh
    vbshvegsh[vbshvegsh > 0] = 1
    vbshvegsh = vbshvegsh-vegsh

    vegsh[vegsh > 0] = 1
    shvoveg = (shvoveg-a) * vegsh    #Vegetation shadow volume
    vegsh = 1-vegsh
    vbshvegsh = 1-vbshvegsh
    #print(np.max(shvoveg))
    wallsh, wallsun, wallshve, facesh, facesun = shade_on_walls(azimuth, aspect, walls, a, f, shvoveg)
    #print(np.max(wallshve))
    if walls_scheme is not False:
        wallsh_, wallsun_, wallshve_, facesh_, facesun_ = shade_on_walls(azimuth, aspect_scheme, walls_scheme, a, f, shvoveg)
        #print(np.max(wallshve_))
        shade_on_wall = wallsh_.copy()
        shade_on_wall[shade_on_wall < wallshve_] = wallshve_[shade_on_wall < wallshve_]

    #return vegsh, sh, vbshvegsh, wallsh, wallsun, wallshve, facesh, facesun, shade_on_wall
    return (vegsh, sh, vbshvegsh, wallsh, wallsun, wallshve, facesh, facesun, shade_on_wall) if walls_scheme is not False else (vegsh, sh, vbshvegsh, wallsh, wallsun, wallshve, facesh, facesun)


import datetime
import logging
import zipfile
from dataclasses import dataclass
from typing import Any, Optional, Union

import numpy as np
import scipy.ndimage as ndi

from .util.SEBESOLWEIGCommonFiles.clearnessindex_2013b import clearnessindex_2013b

# Attempt to import pandas for DataFrame support
# If pandas is not available in QGIS... set it to None
try:
    import pandas as pd
except ImportError:
    pd = None

from . import common
from .functions import wallalgorithms as wa
from .functions.SOLWEIGpython.wall_surface_temperature import load_walls
from .util.SEBESOLWEIGCommonFiles import sun_position as sp

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


@dataclass
class SolweigConfig:
    """Configuration class for SOLWEIG parameters."""

    output_dir: Optional[str] = None
    working_dir: Optional[str] = None
    dsm_path: Optional[str] = None
    svf_path: Optional[str] = None
    wh_path: Optional[str] = None
    wa_path: Optional[str] = None
    use_epw_file: bool = False
    epw_path: Optional[str] = None
    epw_start_date: Optional[str] = None
    epw_end_date: Optional[str] = None
    epw_hours: Optional[str] = None
    met_path: Optional[str] = None
    cdsm_path: Optional[str] = None
    tdsm_path: Optional[str] = None
    dem_path: Optional[str] = None
    lc_path: Optional[str] = None
    aniso_path: Optional[str] = None
    poi_path: Optional[str] = None
    poi_field: Optional[str] = None
    wall_path: Optional[str] = None
    woi_path: Optional[str] = None
    woi_field: Optional[str] = None
    only_global: bool = True
    use_veg_dem: bool = True
    conifer: bool = False
    person_cylinder: bool = True
    utc: int = 0
    use_landcover: bool = True
    use_dem_for_buildings: bool = False
    use_aniso: bool = False
    use_wall_scheme: bool = False
    wall_type: Optional[str] = "Brick"
    output_tmrt: bool = True
    output_kup: bool = True
    output_kdown: bool = True
    output_lup: bool = True
    output_ldown: bool = True
    output_sh: bool = True
    save_buildings: bool = True
    output_kdiff: bool = True
    output_tree_planter: bool = True
    wall_netcdf: bool = False
    plot_poi_patches: bool = False

    def to_file(self, file_path: str):
        """Save configuration to a file."""
        logger.info("Saving configuration to %s", file_path)
        with open(file_path, "w") as f:
            for key in type(self).__annotations__:
                value = getattr(self, key)
                if value is None:
                    value = ""  # Default to empty string if None
                if type(self).__annotations__[key] == bool:
                    f.write(f"{key}={int(value)}\n")
                else:
                    f.write(f"{key}={value}\n")

    def from_file(self, config_path_str: str):
        """Load configuration from a file."""
        config_path = common.check_path(config_path_str)
        logger.info("Loading configuration from %s", config_path)
        with open(config_path) as f:
            for line in f:
                if "=" in line:
                    key, value = line.strip().split("=", 1)
                    if key in type(self).__annotations__:
                        if value.strip() == "":
                            value = None
                        if type(self).__annotations__[key] == bool:
                            setattr(self, key, value == "1" or value.lower() == "true")
                        else:
                            setattr(self, key, value)
                    else:
                        logger.warning("Unknown key in config: %s", key)

    def validate(self):
        """Validate configuration parameters."""
        logger.info("Validating SOLWEIG configuration.")
        if not self.output_dir:
            logger.error("Output directory must be set.")
            raise ValueError("Output directory must be set.")
        self.output_dir = str(common.check_path(self.output_dir, make_dir=True))
        if not self.working_dir:
            logger.error("Working directory must be set.")
            raise ValueError("Working directory must be set.")
        self.working_dir = str(common.check_path(self.working_dir, make_dir=True))
        if not self.dsm_path:
            logger.error("DSM path must be set.")
            raise ValueError("DSM path must be set.")
        if not isinstance(self.utc, int):
            try:
                self.utc = int(self.utc)
            except ValueError as err:
                raise ValueError("UTC offset must be an integer.") from err
        if (self.met_path is None and self.epw_path is None) or (self.met_path and self.epw_path):
            logger.error("Provide either MET or EPW weather file.")
            raise ValueError("Provide either MET or EPW weather file.")
        if self.epw_path is not None:
            if self.epw_start_date is None or self.epw_end_date is None:
                logger.error("EPW start and end dates must be provided if EPW path is set.")
                raise ValueError("EPW start and end dates must be provided if EPW path is set.")
            try:
                if isinstance(self.epw_start_date, str):
                    self.epw_start_date = [int(x) for x in self.epw_start_date.split(",")]
                if isinstance(self.epw_end_date, str):
                    self.epw_end_date = [int(x) for x in self.epw_end_date.split(",")]
                if len(self.epw_start_date) != 4 or len(self.epw_end_date) != 4:
                    logger.error("EPW start and end dates must be in the format: year,month,day,hour")
                    raise ValueError("EPW start and end dates must be in the format: year,month,day,hour")
            except ValueError as err:
                logger.error("Invalid EPW date format: %s or %s", self.epw_start_date, self.epw_end_date)
                raise ValueError(f"Invalid EPW date format: {self.epw_start_date} or {self.epw_end_date}") from err
            if self.epw_hours is None:
                self.epw_hours = list(range(24))
            elif isinstance(self.epw_hours, str):
                self.epw_hours = [int(h) for h in self.epw_hours.split(",")]
            if not all(0 <= h < 24 for h in self.epw_hours):
                logger.error("EPW hours must be between 0 and 23.")
                raise ValueError("EPW hours must be between 0 and 23.")
        if self.use_landcover and self.lc_path is None:
            logger.error("Land cover path must be set if use_landcover is True.")
            raise ValueError("Land cover path must be set if use_landcover is True.")
        if self.use_dem_for_buildings and self.dem_path is None:
            logger.error("DEM path must be set if use_dem_for_buildings is True.")
            raise ValueError("DEM path must be set if use_dem_for_buildings is True.")
        if not self.use_landcover and not self.use_dem_for_buildings:
            logger.error("Either use_landcover or use_dem_for_buildings must be True.")
            raise ValueError("Either use_landcover or use_dem_for_buildings must be True.")
        if self.use_aniso and self.aniso_path is None:
            logger.error("Anisotropic sky path must be set if use_aniso is True.")
            raise ValueError("Anisotropic sky path must be set if use_aniso is True.")
        if self.use_wall_scheme and self.wall_path is None:
            logger.error("Wall scheme path must be set if use_wall_scheme is True.")
            raise ValueError("Wall scheme path must be set if use_wall_scheme is True.")
        if self.plot_poi_patches and (not self.use_aniso or not self.poi_path):
            logger.error("POI path and use_aniso must be set if plot_poi_patches is True.")
            raise ValueError("POI path and use_aniso must be set if plot_poi_patches is True.")


@dataclass
class EnvironData:
    """Class to handle weather data loading and processing."""

    YYYY: np.ndarray
    DOY: np.ndarray
    hours: np.ndarray
    minu: np.ndarray
    Ta: np.ndarray
    RH: np.ndarray
    radG: np.ndarray
    radD: np.ndarray
    radI: np.ndarray
    P: np.ndarray
    Ws: np.ndarray
    altitude: np.ndarray
    azimuth: np.ndarray
    zen: np.ndarray
    jday: np.ndarray
    leafon: np.ndarray
    psi: np.ndarray
    dectime: np.ndarray
    altmax: np.ndarray
    Twater: np.ndarray
    CI: np.ndarray

    def __init__(
        self,
        model_configs: SolweigConfig,
        model_params,
        YYYY: np.ndarray,
        DOY: np.ndarray,
        hours: np.ndarray,
        minu: np.ndarray,
        Ta: np.ndarray,
        RH: np.ndarray,
        radG: np.ndarray,
        radD: np.ndarray,
        radI: np.ndarray,
        P: np.ndarray,
        Ws: np.ndarray,
        location: dict | None,
        UTC: int = 0,
    ):
        """
        This function is used to process the input meteorological file.
        It also calculates Sun position based on the time specified in the met-file
        """
        if location is None:
            raise ValueError("Location must be set before loading MET data.")
        # Initialize attributes
        self.YYYY = YYYY
        self.DOY = DOY
        self.hours = hours
        self.minu = minu
        self.Ta = Ta
        self.RH = RH
        self.radG = radG
        self.radD = radD
        self.radI = radI
        self.P = P
        self.Ws = Ws
        # Calculate remaining attributes
        data_len = len(self.YYYY)
        self.dectime = self.DOY + self.hours / 24 + self.minu / (60 * 24.0)
        if data_len == 1:
            halftimestepdec = 0.0
        else:
            halftimestepdec = float((self.dectime[1] - self.dectime[0]) / 2.0)  # convert from f32
        time = {
            "sec": 0,
            "UTC": UTC,
        }
        sunmaximum = 0.0

        # initialize arrays
        self.altitude = np.empty(data_len, dtype=np.float32)
        self.azimuth = np.empty(data_len, dtype=np.float32)
        self.zen = np.empty(data_len, dtype=np.float32)
        self.jday = np.empty(data_len, dtype=np.float32)
        self.leafon = np.empty(data_len, dtype=np.float32)
        self.psi = np.empty(data_len, dtype=np.float32)
        self.altmax = np.empty(data_len, dtype=np.float32)
        self.Twater = np.empty(data_len, dtype=np.float32)
        self.CI = np.empty(data_len, dtype=np.float32)

        sunmax = dict()

        # These variables lag across days until updated
        Twater = None
        CI = 1

        def _scalar(value: Any) -> float:
            """Convert sun-position outputs to native float scalars."""
            if isinstance(value, (int, float, np.floating)):
                return float(value)
            arr = np.asarray(value)
            if arr.ndim == 0:
                return float(arr)
            return float(arr.reshape(-1)[0])

        # Iterate over time steps and set vars
        for i in range(data_len):
            YMD = datetime.datetime(int(self.YYYY[i]), 1, 1) + datetime.timedelta(int(self.DOY[i]) - 1)
            # Finding maximum altitude in 15 min intervals (20141027)
            if (i == 0) or (np.mod(self.dectime[i], np.floor(self.dectime[i])) == 0):
                fifteen = 0.0
                sunmaximum = -90.0
                sunmax["zenith"] = 90.0
                while sunmaximum <= 90.0 - _scalar(sunmax["zenith"]):
                    sunmaximum = 90.0 - _scalar(sunmax["zenith"])
                    fifteen = fifteen + 15.0 / 1440.0
                    HM = datetime.timedelta(days=(60 * 10) / 1440.0 + fifteen)
                    YMDHM = YMD + HM
                    time["year"] = YMDHM.year
                    time["month"] = YMDHM.month
                    time["day"] = YMDHM.day
                    time["hour"] = YMDHM.hour
                    time["min"] = YMDHM.minute
                    sunmax = sp.sun_position(time, location)
            self.altmax[i] = float(sunmaximum)
            # Calculate sun position
            half = datetime.timedelta(days=halftimestepdec)
            H = datetime.timedelta(hours=int(self.hours[i]))
            M = datetime.timedelta(minutes=int(self.minu[i]))
            YMDHM = YMD + H + M - half
            time["year"] = YMDHM.year
            time["month"] = YMDHM.month
            time["day"] = YMDHM.day
            time["hour"] = YMDHM.hour
            time["min"] = YMDHM.minute
            sun = sp.sun_position(time, location)
            sun_zenith = _scalar(sun["zenith"])
            if (sun_zenith > 89.0) & (
                sun_zenith <= 90.0
            ):  # Hopefully fixes weird values in Perez et al. when altitude < 1.0, i.e. close to sunrise/sunset
                sun_zenith = 89.0
            sun_azimuth = _scalar(sun["azimuth"])
            self.altitude[i] = 90.0 - sun_zenith
            self.zen[i] = sun_zenith * (np.pi / 180.0)
            self.azimuth[i] = sun_azimuth
            # day of year and check for leap year
            # if calendar.isleap(time["year"]):
            #     dayspermonth = np.atleast_2d([31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31])
            # else:
            #     dayspermonth = np.atleast_2d([31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31])
            # jday[0, i] = np.sum(dayspermonth[0, 0:time['month']-1]) + time['day'] # bug when a new day 20191015
            doy = YMD.timetuple().tm_yday
            self.jday[i] = doy
            # Leaf on/off
            if model_configs.conifer:
                # Conifer trees are always leaf on
                self.leafon[i] = 1
            else:
                # Deciduous trees
                self.leafon[i] = 0
                # Check leaf on period
                if model_params.Tree_settings.Value.First_day_leaf > model_params.Tree_settings.Value.Last_day_leaf:
                    self.leafon[i] = int(
                        (model_params.Tree_settings.Value.First_day_leaf < doy)
                        | (model_params.Tree_settings.Value.Last_day_leaf > doy)
                    )
                else:
                    self.leafon[i] = int(
                        (model_params.Tree_settings.Value.First_day_leaf < doy)
                        & (model_params.Tree_settings.Value.Last_day_leaf > doy)
                    )
            # Check if the current time is the start of a new day
            if (self.dectime[i] - np.floor(self.dectime[i])) == 0 or (i == 0):
                # Find average temperature for the current day
                Twater = np.mean(self.Ta[self.jday == np.floor(self.dectime[i])])
            # Lags across hours until updated
            self.Twater[i] = Twater

            # Nocturnal cloudfraction from Offerle et al. 2003
            # Check for start of day
            if (self.dectime[i] - np.floor(self.dectime[i])) == 0:
                # Fallback
                CI = 1.0
                # Find all current day idxs
                daylines = np.where(np.floor(self.dectime) == self.dectime[i])
                # np.where returns a tuple, so check the first element
                if len(daylines[0]) > 1:
                    # Get the altitudes for day's idxs
                    alt_day = self.altitude[daylines[0]]
                    # Find all idxs with altitude greater than 1
                    alt2 = np.where(alt_day > 1)
                    # np.where returns a tuple, so check the first element
                    if len(alt2[0]) > 0:
                        # Take the first altitude greater than 1
                        rise = alt2[0][0]
                        # Calculate clearness index for the next time step after sunrise
                        [_, CI_candidate, _, _, _] = clearnessindex_2013b(
                            self.zen[i + rise + 1],
                            self.jday[i + rise + 1],
                            self.Ta[i + rise + 1],
                            self.RH[i + rise + 1] / 100.0,
                            self.radG[i + rise + 1],
                            location,
                            self.P[i + rise + 1],
                        )
                        if np.isfinite(CI_candidate) and CI_candidate <= 1.0:
                            CI = CI_candidate
            # Lags across hours until updated
            self.CI[i] = CI
        # Calculate psi (transmissivity)
        self.psi = self.leafon * model_params.Tree_settings.Value.Transmissivity
        # TODO: check if this is correct
        self.psi[self.leafon == 0] = 0.5


class SvfData:
    """Class to handle SVF data loading and processing."""

    svf: np.ndarray
    svf_east: np.ndarray
    svf_south: np.ndarray
    svf_west: np.ndarray
    svf_north: np.ndarray
    svf_veg: np.ndarray
    svf_veg_east: np.ndarray
    svf_veg_south: np.ndarray
    svf_veg_west: np.ndarray
    svf_veg_north: np.ndarray
    svf_veg_blocks_bldg_sh: np.ndarray
    svf_veg_blocks_bldg_sh_east: np.ndarray
    svf_veg_blocks_bldg_sh_south: np.ndarray
    svf_veg_blocks_bldg_sh_west: np.ndarray
    svf_veg_blocks_bldg_sh_north: np.ndarray
    svfalfa: np.ndarray

    def __init__(self, model_configs: SolweigConfig):
        logger.info("Loading SVF data from %s", model_configs.svf_path)
        svf_path_str = str(common.check_path(model_configs.svf_path, make_dir=False))
        in_path_str = str(common.check_path(model_configs.working_dir, make_dir=False))
        # Unzip
        with zipfile.ZipFile(svf_path_str, "r") as zip_ref:
            zip_ref.extractall(in_path_str)
        # Load SVF rasters
        self.svf, _, _, _ = common.load_raster(in_path_str + "/" + "svf.tif", coerce_f64_to_f32=True)
        self.svf_east, _, _, _ = common.load_raster(in_path_str + "/" + "svfE.tif", coerce_f64_to_f32=True)
        self.svf_south, _, _, _ = common.load_raster(in_path_str + "/" + "svfS.tif", coerce_f64_to_f32=True)
        self.svf_west, _, _, _ = common.load_raster(in_path_str + "/" + "svfW.tif", coerce_f64_to_f32=True)
        self.svf_north, _, _, _ = common.load_raster(in_path_str + "/" + "svfN.tif", coerce_f64_to_f32=True)
        if model_configs.use_veg_dem:
            self.svf_veg, _, _, _ = common.load_raster(in_path_str + "/" + "svfveg.tif", coerce_f64_to_f32=True)
            self.svf_veg_east, _, _, _ = common.load_raster(in_path_str + "/" + "svfEveg.tif", coerce_f64_to_f32=True)
            self.svf_veg_south, _, _, _ = common.load_raster(in_path_str + "/" + "svfSveg.tif", coerce_f64_to_f32=True)
            self.svf_veg_west, _, _, _ = common.load_raster(in_path_str + "/" + "svfWveg.tif", coerce_f64_to_f32=True)
            self.svf_veg_north, _, _, _ = common.load_raster(in_path_str + "/" + "svfNveg.tif", coerce_f64_to_f32=True)
            self.svf_veg_blocks_bldg_sh, _, _, _ = common.load_raster(
                in_path_str + "/" + "svfaveg.tif", coerce_f64_to_f32=True
            )
            self.svf_veg_blocks_bldg_sh_east, _, _, _ = common.load_raster(
                in_path_str + "/" + "svfEaveg.tif", coerce_f64_to_f32=True
            )
            self.svf_veg_blocks_bldg_sh_south, _, _, _ = common.load_raster(
                in_path_str + "/" + "svfSaveg.tif", coerce_f64_to_f32=True
            )
            self.svf_veg_blocks_bldg_sh_west, _, _, _ = common.load_raster(
                in_path_str + "/" + "svfWaveg.tif", coerce_f64_to_f32=True
            )
            self.svf_veg_blocks_bldg_sh_north, _, _, _ = common.load_raster(
                in_path_str + "/" + "svfNaveg.tif", coerce_f64_to_f32=True
            )
            logger.info("Vegetation SVF data loaded.")
        else:
            self.svf_veg = np.ones_like(self.svf)
            self.svf_veg_east = np.ones_like(self.svf)
            self.svf_veg_south = np.ones_like(self.svf)
            self.svf_veg_west = np.ones_like(self.svf)
            self.svf_veg_north = np.ones_like(self.svf)
            self.svf_veg_blocks_bldg_sh = np.ones_like(self.svf)
            self.svf_veg_blocks_bldg_sh_east = np.ones_like(self.svf)
            self.svf_veg_blocks_bldg_sh_south = np.ones_like(self.svf)
            self.svf_veg_blocks_bldg_sh_west = np.ones_like(self.svf)
            self.svf_veg_blocks_bldg_sh_north = np.ones_like(self.svf)
        # Calculate SVF alpha
        tmp = self.svf + self.svf_veg - 1.0
        tmp[tmp < 0.0] = 0.0
        eps = np.finfo(np.float32).tiny
        safe_term = np.clip(1.0 - tmp, eps, 1.0)
        self.svfalfa = np.arcsin(np.exp(np.log(safe_term) / 2.0))
        logger.info("SVF data loaded and processed.")


def raster_preprocessing(
    dsm: np.ndarray,
    dem: np.ndarray | None,
    cdsm: np.ndarray | None,
    tdsm: np.ndarray | None,
    trunk_ratio: float,
    pix_size: float,
    amax_local_window_m: int = 100,
    amax_local_perc: float = 99.9,
):
    nan32 = np.float32(np.nan)
    zero32 = np.float32(0.0)
    threshold = np.float32(0.1)
    # amax
    if dem is None:
        amaxvalue = float(np.nanmax(dsm) - np.nanmin(dsm))
    else:
        # Calculate local maxima/minima ranges
        # Number of pixels to cover ~amax_local_window_m radius (use a square window)
        radius_pix = max(1, int(np.ceil(amax_local_window_m / pix_size)))
        window = 2 * radius_pix + 1
        try:
            local_min = ndi.minimum_filter(dsm, size=window, mode="nearest")
            local_range = dsm - local_min
            amaxvalue = float(np.nanpercentile(local_range, amax_local_perc))
            logger.info(
                f"amax {amaxvalue}m derived from {amax_local_window_m}m window and {amax_local_perc}th percentile."
            )
        except Exception:
            # Fallback to global range if filtering fails for any reason
            amaxvalue = float(np.nanmax(dsm) - np.nanmin(dsm))
            logger.warning(f"Failed to calculate local amax; using global range of {amaxvalue}m instead.")

    # CDSM is relative to flat surface without DEM
    if cdsm is None:
        cdsm = np.zeros_like(dsm)
        tdsm = np.zeros_like(dsm)
    else:
        if np.nanmax(cdsm) > 50:
            logger.warning(
                f"CDSM max {np.nanmax(cdsm)} exceeds 50 m, vegetation heights to be relative to ground (no DEM)."
            )
        cdsm[np.isnan(cdsm)] = 0.0
        # TDSM is relative to flat surface without DEM
        if tdsm is None:
            logger.info("Tree trunk TDSM not provided; using trunk ratio for TDSM.")
            tdsm = cdsm * trunk_ratio
        if np.nanmax(tdsm) > 50:
            logger.warning(
                f"TDSM max {np.nanmax(tdsm)} exceeds 50m, vegetation heights to be relative to ground (no DEM)."
            )
        if np.nanmax(tdsm) > np.nanmax(cdsm):
            logger.warning("Found TDSM heights exceeding CDSM heights, check input rasters.")
        tdsm[np.isnan(tdsm)] = 0.0

        # Compare veg heights against DSM and update amax if necessary
        # Do before boosting to DEM / CDSM
        vegmax = np.nanmax(cdsm) - np.nanmin(cdsm)
        vegmax = min(vegmax, 50)
        if vegmax > amaxvalue:
            logger.warning(f"Overriding amax {amaxvalue}m with veg max height of {vegmax}m.")
            amaxvalue = vegmax

        # Set vegetated pixels to DEM + CDSM otherwise DSM + CDSM
        if dem is not None:
            cdsm = np.where(~np.isnan(dem), dem + cdsm, nan32)
            cdsm = np.where(cdsm - dem < threshold, zero32, cdsm)
            tdsm = np.where(~np.isnan(dem), dem + tdsm, nan32)
            tdsm = np.where(tdsm - dem < threshold, zero32, tdsm)
        else:
            cdsm = np.where(~np.isnan(dsm), dsm + cdsm, nan32)
            cdsm = np.where(cdsm - dsm < threshold, zero32, cdsm)
            tdsm = np.where(~np.isnan(dsm), dsm + tdsm, nan32)
            tdsm = np.where(tdsm - dsm < threshold, zero32, tdsm)

    logger.info("Calculated max height for shadows: %.2fm", amaxvalue)
    if amaxvalue > 100:
        logger.warning("Max shadow height exceeds 100m, double-check the input rasters for anomalies.")

    dsm = np.asarray(dsm, dtype=np.float32)
    dem = None if dem is None else np.asarray(dem, dtype=np.float32)
    cdsm = np.asarray(cdsm, dtype=np.float32)
    tdsm = np.asarray(tdsm, dtype=np.float32)

    return dsm, dem, cdsm, tdsm, amaxvalue


class RasterData:
    """Class to represent vegetation parameters."""

    amaxvalue: float
    dsm: np.ndarray
    crs_wkt: Optional[str]
    trf_arr: list[float]
    nd_val: Optional[float]
    scale: float
    rows: int
    cols: int
    wallheight: np.ndarray
    wallaspect: np.ndarray
    dem: Optional[np.ndarray]
    cdsm: Optional[np.ndarray]
    tdsm: Optional[np.ndarray]
    bush: np.ndarray
    svfbuveg: np.ndarray
    lcgrid: Optional[np.ndarray]
    buildings: Optional[np.ndarray]

    def __init__(
        self,
        model_configs: SolweigConfig,
        model_params,
        svf_data: SvfData,
        amax_local_window_m: int = 100,
        amax_local_perc: float = 99.9,
    ):
        if model_configs.dsm_path is None:
            raise ValueError("DSM path must be provided before initializing raster data.")
        if model_configs.wh_path is None or model_configs.wa_path is None:
            raise ValueError("Wall height and aspect rasters must be provided.")
        if model_configs.output_dir is None:
            raise ValueError("Output directory must be configured before initializing raster data.")
        output_dir = model_configs.output_dir
        dsm_path = model_configs.dsm_path
        wh_path = model_configs.wh_path
        wa_path = model_configs.wa_path
        # Load DSM
        self.dsm, self.trf_arr, self.crs_wkt, self.nd_val = common.load_raster(
            dsm_path, bbox=None, coerce_f64_to_f32=True
        )
        logger.info("DSM loaded from %s", dsm_path)
        self.scale = 1 / self.trf_arr[1]
        self.rows = self.dsm.shape[0]
        self.cols = self.dsm.shape[1]
        one32 = np.float32(1.0)
        zero32 = np.float32(0.0)
        self.dsm = np.ascontiguousarray(self.dsm, dtype=np.float32)
        # TODO: is this needed?
        # if self.dsm.min() < 0:
        #     dsmraise = np.abs(self.dsm.min())
        #     self.dsm = self.dsm + dsmraise
        # else:
        #     dsmraise = 0

        # WALLS
        # heights
        self.wallheight, wh_trf, wh_crs, _ = common.load_raster(wh_path, bbox=None, coerce_f64_to_f32=True)
        if not self.wallheight.shape == self.dsm.shape:
            raise ValueError("Mismatching raster shapes for wall heights and DSM.")
        if not np.allclose(self.trf_arr, wh_trf):
            raise ValueError("Mismatching spatial transform for wall heights and DSM.")
        if not self.crs_wkt == wh_crs:
            raise ValueError("Mismatching CRS for wall heights and DSM.")
        logger.info("Wall heights loaded")
        self.wallheight = np.ascontiguousarray(self.wallheight, dtype=np.float32)
        # aspects
        self.wallaspect, wa_trf, wa_crs, _ = common.load_raster(wa_path, bbox=None, coerce_f64_to_f32=True)
        if not self.wallaspect.shape == self.dsm.shape:
            raise ValueError("Mismatching raster shapes for wall aspects and DSM.")
        if not np.allclose(self.trf_arr, wa_trf):
            raise ValueError("Mismatching spatial transform for wall aspects and DSM.")
        if not self.crs_wkt == wa_crs:
            raise ValueError("Mismatching CRS for wall aspects and DSM.")
        logger.info("Wall aspects loaded")
        self.wallaspect = np.ascontiguousarray(self.wallaspect, dtype=np.float32)

        # DEM
        # TODO: Is DEM always provided?
        if model_configs.dem_path:
            dem_path_str = str(common.check_path(model_configs.dem_path))
            self.dem, dem_trf, dem_crs, dem_nd_val = common.load_raster(dem_path_str, bbox=None, coerce_f64_to_f32=True)
            if not self.dem.shape == self.dsm.shape:
                raise ValueError("Mismatching raster shapes for DEM and CDSM.")
            if dem_crs is not None and dem_crs != self.crs_wkt:
                raise ValueError("Mismatching CRS for DEM and CDSM.")
            if not np.allclose(self.trf_arr, dem_trf):
                raise ValueError("Mismatching spatial transform for DEM and CDSM.")
            logger.info("DEM loaded from %s", model_configs.dem_path)
            self.dem = np.ascontiguousarray(self.dem, dtype=np.float32)
            # dem[dem == dem_nd_val] = 0.0
            # TODO: Check if this is needed re DSM ramifications
            # if dem.min() < 0:
            #     demraise = np.abs(dem.min())
            #     dem = dem + demraise
        else:
            self.dem = None

        # Vegetation
        if model_configs.use_veg_dem:
            self.cdsm, vegdsm_trf, vegdsm_crs, _ = common.load_raster(
                model_configs.cdsm_path, bbox=None, coerce_f64_to_f32=True
            )
            if not self.cdsm.shape == self.dsm.shape:
                raise ValueError("Mismatching raster shapes for DSM and CDSM.")
            if vegdsm_crs is not None and vegdsm_crs != self.crs_wkt:
                raise ValueError("Mismatching CRS for DSM and CDSM.")
            if not np.allclose(self.trf_arr, vegdsm_trf):
                raise ValueError("Mismatching spatial transform for DSM and CDSM.")
            logger.info("Vegetation DSM loaded from %s", model_configs.cdsm_path)
            self.cdsm = np.ascontiguousarray(self.cdsm, dtype=np.float32)
            # Tree DSM
            if model_configs.tdsm_path:
                self.tdsm, vegdsm2_trf, vegdsm2_crs, _ = common.load_raster(
                    model_configs.tdsm_path, bbox=None, coerce_f64_to_f32=True
                )
                if not self.tdsm.shape == self.dsm.shape:
                    raise ValueError("Mismatching raster shapes for DSM and CDSM.")
                if vegdsm2_crs is not None and vegdsm2_crs != self.crs_wkt:
                    raise ValueError("Mismatching CRS for DSM and CDSM.")
                if not np.allclose(self.trf_arr, vegdsm2_trf):
                    raise ValueError("Mismatching spatial transform for DSM and CDSM.")
                logger.info("Tree DSM loaded from %s", model_configs.tdsm_path)
                self.tdsm = np.ascontiguousarray(self.tdsm, dtype=np.float32)
            else:
                self.tdsm = None
        else:
            self.cdsm = None
            self.tdsm = None

        self.dsm, self.dem, self.cdsm, self.tdsm, self.amaxvalue = raster_preprocessing(  # type: ignore
            self.dsm,
            self.dem,
            self.cdsm,
            self.tdsm,
            model_params.Tree_settings.Value.Trunk_ratio,
            self.trf_arr[1],
            amax_local_window_m,
            amax_local_perc,
        )
        self.dsm = np.ascontiguousarray(self.dsm, dtype=np.float32)
        if self.dem is not None:
            self.dem = np.ascontiguousarray(self.dem, dtype=np.float32)
        if self.cdsm is not None:
            self.cdsm = np.ascontiguousarray(self.cdsm, dtype=np.float32)
        if self.tdsm is not None:
            self.tdsm = np.ascontiguousarray(self.tdsm, dtype=np.float32)

        # bushes etc
        if model_configs.use_veg_dem:
            transmissivity = np.float32(model_params.Tree_settings.Value.Transmissivity)
            self.bush = np.logical_not(self.tdsm * self.cdsm) * self.cdsm
            self.bush = np.ascontiguousarray(self.bush, dtype=np.float32)
            self.svfbuveg = svf_data.svf - (one32 - svf_data.svf_veg) * (one32 - transmissivity)
            self.svfbuveg = np.ascontiguousarray(self.svfbuveg, dtype=np.float32)
        else:
            logger.info("Vegetation DEM not used; vegetation arrays set to None.")
            self.bush = np.zeros([self.rows, self.cols], dtype=np.float32)
            self.svfbuveg = np.ascontiguousarray(svf_data.svf, dtype=np.float32)

        common.save_raster(
            output_dir + "/input-dsm.tif",
            self.dsm,
            self.trf_arr,
            self.crs_wkt,
            self.nd_val,
            coerce_f64_to_f32=True,
        )
        common.save_raster(
            output_dir + "/input-cdsm.tif",
            self.cdsm,
            self.trf_arr,
            self.crs_wkt,
            self.nd_val,
            coerce_f64_to_f32=True,
        )
        common.save_raster(
            output_dir + "/input-tdsm.tif",
            self.tdsm,
            self.trf_arr,
            self.crs_wkt,
            self.nd_val,
            coerce_f64_to_f32=True,
        )
        common.save_raster(
            output_dir + "/input-svfbuveg.tif",
            self.svfbuveg,
            self.trf_arr,
            self.crs_wkt,
            self.nd_val,
            coerce_f64_to_f32=True,
        )
        common.save_raster(
            output_dir + "/input-bush.tif",
            self.bush,
            self.trf_arr,
            self.crs_wkt,
            self.nd_val,
            coerce_f64_to_f32=True,
        )

        # Land cover
        if model_configs.use_landcover:
            lc_path = model_configs.lc_path
            if lc_path is None:
                raise ValueError("Land cover path must be provided when use_landcover is True.")
            lc_path_str = str(common.check_path(lc_path))
            self.lcgrid, lc_trf, lc_crs, _ = common.load_raster(lc_path_str, bbox=None, coerce_f64_to_f32=True)
            if not self.lcgrid.shape == self.dsm.shape:
                raise ValueError("Mismatching raster shapes for land cover and DSM.")
            if lc_crs is not None and lc_crs != self.crs_wkt:
                raise ValueError("Mismatching CRS for land cover and DSM.")
            if not np.allclose(self.trf_arr, lc_trf):
                raise ValueError("Mismatching spatial transform for land cover and DSM.")
            logger.info("Land cover loaded from %s", lc_path)
            self.lcgrid = np.ascontiguousarray(self.lcgrid, dtype=np.float32)
        else:
            self.lcgrid = None
            logger.info("Land cover not used; lcgrid set to None.")

        # Buildings from land cover option
        # TODO: Check intended logic here
        if not model_configs.use_dem_for_buildings and self.lcgrid is not None:
            # Create building boolean raster from either land cover if no DEM is used
            buildings = np.copy(self.lcgrid)
            buildings[buildings == 7] = 1
            buildings[buildings == 6] = 1
            buildings[buildings == 5] = 1
            buildings[buildings == 4] = 1
            buildings[buildings == 3] = 1
            buildings[buildings == 2] = 0
            self.buildings = np.ascontiguousarray(buildings, dtype=np.float32)
            logger.info("Buildings raster created from land cover data.")
        elif model_configs.use_dem_for_buildings:
            if self.dem is None:
                raise ValueError("DEM raster must be available when use_dem_for_buildings is True.")
            height_diff = self.dsm - self.dem
            buildings = np.where(
                ~np.isnan(self.dem) & ~np.isnan(self.dsm),
                height_diff,
                zero32,
            )
            # TODO: Check intended logic here - 1 vs 0
            buildings[buildings < 2.0] = 1.0
            buildings[buildings >= 2.0] = 0.0
            self.buildings = np.ascontiguousarray(buildings, dtype=np.float32)
            logger.info("Buildings raster created from DSM and DEM data.")
        else:
            self.buildings = None
            logger.info("Buildings raster not created.")
        # Save buildings raster if requested
        if self.buildings is not None and model_configs.save_buildings:
            common.save_raster(
                output_dir + "/buildings.tif",
                self.buildings,
                self.trf_arr,
                self.crs_wkt,
                self.nd_val,
                coerce_f64_to_f32=True,
            )
            logger.info("Buildings raster saved to %s/buildings.tif", output_dir)


class ShadowMatrices:
    """Shadow matrices and related anisotropic sky data."""

    use_aniso: bool
    shmat: Optional[np.ndarray]
    diffsh: Optional[np.ndarray]
    vegshmat: Optional[np.ndarray]
    vbshvegshmat: Optional[np.ndarray]
    asvf: Optional[np.ndarray]
    patch_option: int
    steradians: Union[int, np.ndarray]

    def __init__(
        self,
        model_configs: SolweigConfig,
        model_params,
        svf_data: SvfData,
    ):
        self.use_aniso = model_configs.use_aniso
        if self.use_aniso:
            logger.info("Loading anisotropic shadow matrices from %s", model_configs.aniso_path)
            aniso_path_str = str(common.check_path(model_configs.aniso_path, make_dir=False))
            data = np.load(aniso_path_str)
            self.shmat = data["shadowmat"].astype(np.float32)
            self.vegshmat = data["vegshadowmat"].astype(np.float32)
            self.vbshvegshmat = data["vbshmat"].astype(np.float32)
            if model_configs.use_veg_dem:
                # TODO: thoughts on memory optimization for smaller machines / large arrays?
                self.diffsh = (
                    self.shmat - (1 - self.vegshmat) * (1 - model_params.Tree_settings.Value.Transmissivity)
                ).astype(np.float32)
                """
                self.diffsh = np.zeros((raster_data.rows, raster_data.cols, self.shmat.shape[2]))
                for i in range(0, self.shmat.shape[2]):
                    self.diffsh[:, :, i] = self.shmat[:, :, i] - (1 - self.vegshmat[:, :, i]) * (
                        1 - model_params.Tree_settings.Value.Transmissivity
                    )
                """
                logger.info("Shadow matrices with vegetation loaded.")
            else:
                self.diffsh = self.shmat
                logger.info("Shadow matrices loaded (no vegetation).")

            # Estimate number of patches based on shadow matrices
            if self.shmat.shape[2] == 145:
                self.patch_option = 1  # patch_option = 1 # 145 patches
            elif self.shmat.shape[2] == 153:
                self.patch_option = 2  # patch_option = 2 # 153 patches
            elif self.shmat.shape[2] == 306:
                self.patch_option = 3  # patch_option = 3 # 306 patches
            elif self.shmat.shape[2] == 612:
                self.patch_option = 4  # patch_option = 4 # 612 patches

            # asvf to calculate sunlit and shaded patches
            self.asvf = np.arccos(np.sqrt(svf_data.svf))

            # Empty array for steradians
            self.steradians = np.zeros(self.shmat.shape[2], dtype=np.float32)
        else:
            # no anisotropic sky
            # downstream functions only access these if use_aniso is True
            # be aware that Solweig_2025a_calc expects an int not bool for use_aniso
            self.diffsh = None
            self.shmat = None
            self.vegshmat = None
            self.vbshvegshmat = None
            self.asvf = None
            self.patch_option = 0
            self.steradians = 0
            logger.info("Anisotropic sky not used; shadow matrices not loaded.")


class TgMaps:
    """
    Get land cover properties for Tg wave (land cover scheme based on Bogren et al. 2000,
    explained in Lindberg et al., 2008 and Lindberg, Onomura & Grimmond, 2016)
    """

    TgK: np.ndarray
    Tstart: np.ndarray
    alb_grid: np.ndarray
    emis_grid: np.ndarray
    TgK_wall: float
    Tstart_wall: float
    TmaxLST: Union[np.ndarray, float]
    TmaxLST_wall: float
    Knight: np.ndarray
    Tgmap1: np.ndarray
    Tgmap1E: np.ndarray
    Tgmap1S: np.ndarray
    Tgmap1W: np.ndarray
    Tgmap1N: np.ndarray
    TgOut1: np.ndarray

    def __init__(self, use_landcover: bool, model_params, raster_data: RasterData):
        """
        This is a vectorized version that avoids looping over pixels.
        """
        # Initialization of maps
        self.Knight = np.zeros((raster_data.rows, raster_data.cols), dtype=np.float32)
        self.Tgmap1 = np.zeros((raster_data.rows, raster_data.cols), dtype=np.float32)
        self.Tgmap1E = np.zeros((raster_data.rows, raster_data.cols), dtype=np.float32)
        self.Tgmap1S = np.zeros((raster_data.rows, raster_data.cols), dtype=np.float32)
        self.Tgmap1W = np.zeros((raster_data.rows, raster_data.cols), dtype=np.float32)
        self.Tgmap1N = np.zeros((raster_data.rows, raster_data.cols), dtype=np.float32)
        self.TgOut1 = np.zeros((raster_data.rows, raster_data.cols), dtype=np.float32)

        # Set up the Tg maps based on whether land cover is used
        if use_landcover is False:
            self.TgK = self.Knight + model_params.Ts_deg.Value.Cobble_stone_2014a
            self.Tstart = self.Knight - model_params.Tstart.Value.Cobble_stone_2014a
            self.alb_grid = self.Knight + model_params.Albedo.Effective.Value.Cobble_stone_2014a
            self.emis_grid = self.Knight + model_params.Emissivity.Value.Cobble_stone_2014a
            self.TmaxLST = model_params.TmaxLST.Value.Cobble_stone_2014a  # Assuming this is a float
            self.TgK_wall = model_params.Ts_deg.Value.Walls
            self.Tstart_wall = model_params.Tstart.Value.Walls
            self.TmaxLST_wall = model_params.TmaxLST.Value.Walls
            logger.info("TgMaps initialized with default (no land cover) parameters.")
        else:
            if raster_data.lcgrid is None:
                raise ValueError("Land cover grid is not available.")
            # Copy land cover grid
            lc_grid = np.copy(raster_data.lcgrid)
            # Sanitize
            lc_grid[lc_grid >= 100] = 2
            # Get unique land cover IDs and filter them
            unique_ids = np.unique(lc_grid)
            valid_ids = unique_ids[unique_ids <= 7].astype(int)
            # Initialize output grids by copying the original land cover grid
            self.TgK = np.copy(lc_grid)
            self.Tstart = np.copy(lc_grid)
            self.alb_grid = np.copy(lc_grid)
            self.emis_grid = np.copy(lc_grid)
            self.TmaxLST = np.copy(lc_grid)
            # Create mapping dictionaries from land cover ID to parameter values
            id_to_name = {i: getattr(model_params.Names.Value, str(i)) for i in valid_ids}
            name_to_tstart = {name: getattr(model_params.Tstart.Value, name) for name in id_to_name.values()}
            name_to_albedo = {name: getattr(model_params.Albedo.Effective.Value, name) for name in id_to_name.values()}
            name_to_emissivity = {name: getattr(model_params.Emissivity.Value, name) for name in id_to_name.values()}
            name_to_tmaxlst = {name: getattr(model_params.TmaxLST.Value, name) for name in id_to_name.values()}
            name_to_tsdeg = {name: getattr(model_params.Ts_deg.Value, name) for name in id_to_name.values()}
            # Perform replacements for each valid land cover ID
            for i in valid_ids:
                mask = lc_grid == i
                name = id_to_name[i]
                self.Tstart[mask] = name_to_tstart[name]
                self.alb_grid[mask] = name_to_albedo[name]
                self.emis_grid[mask] = name_to_emissivity[name]
                self.TmaxLST[mask] = name_to_tmaxlst[name]
                self.TgK[mask] = name_to_tsdeg[name]
            # Get wall-specific parameters
            self.TgK_wall = getattr(model_params.Ts_deg.Value, "Walls", None)
            self.Tstart_wall = getattr(model_params.Tstart.Value, "Walls", None)
            self.TmaxLST_wall = getattr(model_params.TmaxLST.Value, "Walls", None)
            logger.info("TgMaps initialized using land cover grid.")


class WallsData:
    """Class to represent wall characteristics and configurations."""

    voxelMaps: Optional[np.ndarray]
    voxelTable: Optional[np.ndarray]
    timeStep: int
    walls_scheme: np.ndarray
    dirwalls_scheme: np.ndarray
    met_for_xarray: Optional[Any]

    def __init__(
        self,
        model_configs: SolweigConfig,
        model_params,
        raster_data: RasterData,
        weather_data: EnvironData,
        tg_maps: TgMaps,
    ):
        if model_configs.use_wall_scheme:
            logger.info("Loading wall scheme data from %s", model_configs.wall_path)
            wall_path_str = str(common.check_path(model_configs.wall_path, make_dir=False))
            wallData = np.load(wall_path_str)
            #
            self.voxelMaps = wallData["voxelId"]
            self.voxelTable = wallData["voxelTable"]
            # Get wall type
            # TODO:
            # wall_type_standalone = {"Brick_wall": "100", "Concrete_wall": "101", "Wood_wall": "102"}
            wall_type = model_configs.wall_type
            # Get heights of walls including corners
            self.walls_scheme = wa.findwalls_sp(raster_data.dsm, 2, np.array([[1, 1, 1], [1, 0, 1], [1, 1, 1]]))
            # Get aspects of walls including corners
            self.dirwalls_scheme = wa.filter1Goodwin_as_aspect_v3(
                self.walls_scheme.copy(), raster_data.scale, raster_data.dsm, None, 100.0 / 180.0
            )
            # Calculate timeStep
            first_timestep = (
                pd.to_datetime(weather_data.YYYY[0], format="%Y")
                + pd.to_timedelta(weather_data.DOY[0] - 1, unit="d")
                + pd.to_timedelta(weather_data.hours[0], unit="h")
                + pd.to_timedelta(weather_data.minu[0], unit="m")
            )
            second_timestep = (
                pd.to_datetime(weather_data.YYYY[1], format="%Y")
                + pd.to_timedelta(weather_data.DOY[1] - 1, unit="d")
                + pd.to_timedelta(weather_data.hours[1], unit="h")
                + pd.to_timedelta(weather_data.minu[1], unit="m")
            )
            self.timeStep = (second_timestep - first_timestep).seconds
            # Load voxelTable as Pandas DataFrame
            self.voxelTable, self.dirwalls_scheme = load_walls(
                self.voxelTable,
                model_params,
                wall_type,
                self.dirwalls_scheme,
                weather_data.Ta[0],
                self.timeStep,
                tg_maps.alb_grid,
                model_configs.use_landcover,
                raster_data.lcgrid,
                raster_data.dsm,
            )
            # Create pandas datetime object for NetCDF output
            self.met_for_xarray = (
                pd.to_datetime(weather_data.YYYY, format="%Y")
                + pd.to_timedelta(weather_data.DOY - 1, unit="d")
                + pd.to_timedelta(weather_data.hours, unit="h")
                + pd.to_timedelta(weather_data.minu, unit="m")
            )
            logger.info("Wall scheme data loaded and processed.")
        else:
            self.voxelMaps = None
            self.voxelTable = None
            self.timeStep = 0
            self.walls_scheme = np.ones((raster_data.rows, raster_data.cols), dtype=np.float32) * 10.0
            self.dirwalls_scheme = np.ones((raster_data.rows, raster_data.cols), dtype=np.float32) * 10.0
            self.met_for_xarray = None
            logger.info("Wall scheme not used; default wall data initialized.")
